# from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, Optional
from core.models.dataset_meta import DatasetMetadata

@dataclass
class DataSet:
    meta: Optional[DatasetMetadata]
    dataframeIds : List["str"] = field(default_factory=list)
    accessible: bool = False
    accessInfo: Optional["dict"] = None


from __future__ import annotations

import pandas
import pyarrow as pa
from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Any


class DataFrame(ABC):

    def __init__(self, id: str):
        self.id = id
        self.data = None
        self.actions = []

    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def collect(self) -> DataFrame:
        pass

    @abstractmethod
    def get_stream(self, max_chunksize: Optional[int] = None):
        pass

    @abstractmethod
    def limit(self, rowNum: int) -> DataFrame:
        pass

    @abstractmethod
    def slice(self, offset: int = 0, length: Optional[int] = None) -> DataFrame:
        pass

    @abstractmethod
    def select(self, *columns):
        pass

    @abstractmethod
    def filter(self, mask: pa.Array) -> DataFrame:
        pass

    @abstractmethod
    def sum(self, column: str):
        pass

    @abstractmethod
    def map(self, column: str, func: Any, new_column_name: Optional[str] = None) -> DataFrame:
        pass

    @abstractmethod
    def to_pandas(self, **kwargs) -> pandas.DataFrame:
        pass

    @abstractmethod
    def to_pydict(self) -> Dict[str, List[Any]]:
        pass

    @abstractmethod
    def to_string(
            self,
            head_rows: int = 5,
            tail_rows: int = 5,
            first_cols: int = 3,
            last_cols: int = 3,
            display_all: bool = False
    ) -> str:
        pass

    @staticmethod
    def from_pandas(pdf: pandas.DataFrame, schema=None) -> DataFrame:
        pass


import re
from datetime import date
from typing import List, Dict, Optional, Union
from pydantic import BaseModel, HttpUrl, Field, field_validator

class Identifier(BaseModel):
    """唯一标识符（多值字典结构）"""
    id: HttpUrl  # 符合URI格式的标识，如DOI或CSTR
    type: str = Field(..., pattern="^(DOI|CSTR|InternalID)$")  # 标识类型约束

class BasicInfo(BaseModel):
    """基本信息"""
    name: Optional[str] = None
    identifier: Optional[List[Identifier]] = None  # 多值字典
    description: Optional[str] = None
    keywords: Optional[List[str]] = None
    url: Optional[HttpUrl] = None
    dateCreated: Optional[date] = None
    subject: Optional[List[str]] = None
    format: Optional[List[str]] = None  # 可选字段
    image: Optional[HttpUrl] = None     # 可选字段

class AccessRights(BaseModel):
    """访问权限信息"""
    type: str  # 如 "open"
    openDate: Optional[date] = None
    condition: Optional[str] = None

class DistributionInfo(BaseModel):
    """分发信息"""
    accessRights: Optional[AccessRights] = None
    license: Optional[str] = None
    byteSize: Optional[float] = None
    fileNumber: Optional[int] = None
    downloadURL: Optional[HttpUrl] = None

class RightsInfo(BaseModel):
    """权益信息"""
    creator: Optional[List[str]] = None
    publisher: Optional[str] = None
    contactPoint: Optional[List[str]] = None
    email: Optional[List[str]] = None
    copyrightHolder: Optional[List[str]] = None
    references: Optional[List[HttpUrl]] = None

    # 邮箱格式验证
    @field_validator('email')
    @classmethod
    def validate_emails(cls, v):
        email_pattern = re.compile(r"^[\w\.-]+@[\w\.-]+\.\w+$")
        for email in v:
            if not email_pattern.match(email):
                raise ValueError(f"Invalid email format: {email}")
        return v

# class InstrumentInfo(BaseModel):
#     """装置信息"""
#     instrumentID: str
#     model: str
#     name: str
#     description: str
#     supportingInstitution: str
#     manufacuturer: str
#     accountablePerson: str
#     contactPoint: str
#     email: List[str]
#
#     # 邮箱格式验证
#     @field_validator('email')
#     @classmethod
#     def validate_emails(cls, v):
#         email_pattern = re.compile(r"^[\w\.-]+@[\w\.-]+\.\w+$")
#         for email in v:
#             if not email_pattern.match(email):
#                 raise ValueError(f"Invalid email format: {email}")
#         return v

# -------------------元数据模型 -------------------
class DatasetMetadata(BaseModel):
    basic: BasicInfo
    distribution: DistributionInfo
    rights: RightsInfo
    #instrument: InstrumentInfo


# ------------------- 使用示例 -------------------
# if __name__ == "__main__":
#     example = DatasetMetadata(
#         basic=BasicInfo(
#             name="2020-2023年青藏高原冰川监测数据集",
#             identifier=[Identifier(id="http://doi.org/10.1234/example", type="DOI")],
#             description="本数据集包含青藏高原主要冰川的年度厚度变化监测数据本数据集包含青藏高原主要冰川的年度厚度变化监测数据本数据集包含青藏高原主要冰川的年度厚度变化监测数据本数据集包含青藏高原主要冰川的年度厚度变化监测数据本数据集包含青藏高原主要冰川的年度厚度变化监测数据...（不少于50字）",
#             keywords=["冰川学", "遥感监测", "气候变化"],
#             url="https://example.com/dataset/123",
#             datePublished=date(2023, 5, 1),
#             subject=["地球科学>冰川学"]
#         ),
#         distribution=DistributionInfo(
#             license="https://creativecommons.org/licenses/by/4.0/"
#         ),
#         rights=RightsInfo(
#             creator=["中国科学院青藏高原研究所"],
#             publisher="国家冰川数据中心",
#             contactPoint=["张研究员"],
#             email=["contact@example.com"]
#         ),
#         instrument=InstrumentInfo(
#             instrumentID="GEO-001",
#             model="GLACIER-2020",
#             name="多光谱冰川监测雷达",
#             description="采用X波段雷达测量冰川厚度...",
#             supportingInstitution="中国科学院",
#             manufacturer="中电科集团",
#             accountablePerson="王主任",
#             contactPoint="李工程师",
#             email=["support@example.com"]
#         )
#     )
#     logger.info(example.model_dump_json(indent=2))


import atexit
import threading
import time
import queue
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Optional, Dict, Any
import pyarrow.flight as fl
import pyarrow as pa
from pyarrow._flight import FlightClient


class FlightConnectionPool:
    """Arrow Flight RPC 连接池实现 - 改进版"""

    def __init__(self,
                 location: str,
                 max_connections: int = 10,
                 min_connections: int = 1,
                 connection_timeout: int = 30,
                 idle_timeout: int = 300,
                 max_wait_time: int = 60,  # 新增：最大等待时间
                 enable_blocking_wait: bool = True,  # 新增：是否启用阻塞等待
                 **connection_kwargs):
        """
        初始化连接池

        Args:
            location: Flight 服务器地址
            max_connections: 最大连接数
            min_connections: 最小连接数
            connection_timeout: 连接超时时间(秒)
            idle_timeout: 连接空闲超时时间(秒)
            max_wait_time: 等待可用连接的最大时间(秒)
            enable_blocking_wait: 是否启用阻塞等待模式
            **connection_kwargs: 传递给 FlightClient 的其他参数
        """
        self.location = location
        self.max_connections = max_connections
        self.min_connections = min_connections
        self.connection_timeout = connection_timeout
        self.idle_timeout = idle_timeout
        self.max_wait_time = max_wait_time
        self.enable_blocking_wait = enable_blocking_wait
        self.connection_kwargs = connection_kwargs

        # 连接池队列
        self._pool = queue.Queue(maxsize=max_connections)
        self._active_connections = {}  # 活跃连接
        self._connection_count = 0
        self._lock = threading.RLock()

        # 新增：等待队列，用于管理等待连接的请求
        self._waiting_queue = queue.Queue()
        self._shutdown = False

        # 初始化最小连接数
        self._initialize_pool()

        # 启动清理线程
        self._cleanup_thread = threading.Thread(target=self._cleanup_idle_connections, daemon=True)
        self._cleanup_thread.start()

    def _create_connection(self) -> fl.FlightClient:
        """创建新的 Flight 连接"""
        try:
            client = fl.connect(self.location, **self.connection_kwargs)
            return client
        except Exception as e:
            raise ConnectionError(f"Failed to create Flight connection: {e}")

    def _initialize_pool(self):
        """初始化连接池"""
        with self._lock:
            for _ in range(self.min_connections):
                try:
                    conn = self._create_connection()
                    self._pool.put({
                        'client': conn,
                        'created_at': time.time(),
                        'last_used': time.time()
                    })
                    self._connection_count += 1
                except Exception as e:
                    print(f"Warning: Failed to initialize connection: {e}")

    def _cleanup_idle_connections(self):
        """清理空闲连接的后台线程"""
        while not self._shutdown:
            try:
                time.sleep(60)  # 每分钟检查一次
                current_time = time.time()

                with self._lock:
                    # 收集需要清理的连接
                    temp_connections = []

                    while not self._pool.empty() and self._connection_count > self.min_connections:
                        try:
                            conn_info = self._pool.get_nowait()
                            if current_time - conn_info['last_used'] > self.idle_timeout:
                                # 连接空闲时间过长，关闭连接
                                try:
                                    conn_info['client'].close()
                                except:
                                    pass
                                self._connection_count -= 1
                            else:
                                # 连接仍然有效，放回队列
                                temp_connections.append(conn_info)
                        except queue.Empty:
                            break

                    # 将有效连接放回队列
                    for conn_info in temp_connections:
                        self._pool.put(conn_info)

            except Exception as e:
                if not self._shutdown:
                    print(f"Error in cleanup thread: {e}")

    def _get_connection(self, timeout: Optional[float] = None) -> fl.FlightClient:
        """
        从连接池获取连接 - 改进版

        Args:
            timeout: 获取连接的超时时间

        Returns:
            FlightClient 实例
        """
        if timeout is None:
            timeout = self.connection_timeout

        start_time = time.time()

        while True:
            try:
                # 尝试从池中获取现有连接
                conn_info = self._pool.get(timeout=0.1)  # 短时间超时
                conn_info['last_used'] = time.time()

                # 将连接标记为活跃
                conn_id = id(conn_info['client'])
                with self._lock:
                    self._active_connections[conn_id] = conn_info

                return conn_info['client']

            except queue.Empty:
                # 检查是否可以创建新连接
                with self._lock:
                    if self._connection_count < self.max_connections:
                        try:
                            client = self._create_connection()
                            self._connection_count += 1

                            conn_info = {
                                'client': client,
                                'created_at': time.time(),
                                'last_used': time.time()
                            }

                            conn_id = id(client)
                            self._active_connections[conn_id] = conn_info

                            return client
                        except Exception as e:
                            print(f"Failed to create new connection: {e}")

                # 如果启用阻塞等待且未超时，继续等待
                if self.enable_blocking_wait:
                    elapsed = time.time() - start_time
                    if elapsed >= timeout:
                        raise ConnectionError(f"Timeout waiting for available connection after {elapsed:.2f}s")

                    # 短暂睡眠后重试
                    time.sleep(0.1)
                else:
                    # 不启用阻塞等待，直接抛出异常
                    raise ConnectionError("Connection pool exhausted and max connections reached")

    def return_connection(self, client: fl.FlightClient):
        """
        将连接返回到连接池

        Args:
            client: 要返回的 FlightClient 实例
        """
        conn_id = id(client)

        with self._lock:
            if conn_id in self._active_connections:
                conn_info = self._active_connections.pop(conn_id)
                conn_info['last_used'] = time.time()

                try:
                    # 检查连接是否仍然有效
                    # 这里可以添加连接健康检查
                    self._pool.put(conn_info, block=False)
                except queue.Full:
                    # 连接池已满，关闭连接
                    try:
                        client.close()
                    except:
                        pass
                    self._connection_count -= 1

    @contextmanager
    def get_client(self, timeout: Optional[float] = None):
        """
        上下文管理器方式获取连接

        Usage:
            with pool.get_client() as client:
                # 使用 client 进行操作
                pass
        """
        client = None
        try:
            client = self._get_connection(timeout=timeout)
            yield client
        finally:
            if client:
                self.return_connection(client)

    def close_all(self):
        """关闭所有连接"""
        self._shutdown = True

        with self._lock:
            # 关闭活跃连接
            for conn_info in self._active_connections.values():
                try:
                    conn_info['client'].close()
                except:
                    pass
            self._active_connections.clear()

            # 关闭池中的连接
            while not self._pool.empty():
                try:
                    conn_info = self._pool.get_nowait()
                    conn_info['client'].close()
                except:
                    pass

            self._connection_count = 0

    def get_stats(self) -> Dict[str, Any]:
        """获取连接池统计信息"""
        with self._lock:
            return {
                'total_connections': self._connection_count,
                'active_connections': len(self._active_connections),
                'available_connections': self._pool.qsize(),
                'max_connections': self.max_connections,
                'min_connections': self.min_connections,
                'waiting_requests': self._waiting_queue.qsize()
            }

    def resize_pool(self, new_max_connections: int):
        """动态调整连接池大小"""
        with self._lock:
            if new_max_connections < self.min_connections:
                raise ValueError("Max connections cannot be less than min connections")

            old_max = self.max_connections
            self.max_connections = new_max_connections

            print(f"Connection pool resized from {old_max} to {new_max_connections}")


class ConnectionManager:
    _pool: Optional[FlightConnectionPool] = None

    @staticmethod
    def set_connection_pool(pool: FlightConnectionPool):
        ConnectionManager._pool = pool

    @staticmethod
    def get_connection(timeout: Optional[float] = None) -> FlightClient:
        if ConnectionManager._pool is None:
            raise RuntimeError("Connection pool has not been initialized.")
        return ConnectionManager._pool.get_client(timeout=timeout)

    @staticmethod
    def close_connection_pool():
        print(f"Closing connection pool: {ConnectionManager._pool}")
        if ConnectionManager._pool:
            ConnectionManager._pool.close_all()
            ConnectionManager._pool = None

    @staticmethod
    def get_pool_stats():
        if ConnectionManager._pool:
            return ConnectionManager._pool.get_stats()
        return None


atexit.register(ConnectionManager.close_connection_pool)


def stress_test_connection_pool():
    """强制测试连接池满的情况"""
    # 创建小容量连接池的客户端
    ConnectionManager.set_connection_pool(FlightConnectionPool("grpc://localhost:8080", max_connections=20, idle_timeout=30))

    print("=== 开始连接池压力测试 ===")
    print("初始连接池状态:", ConnectionManager._pool.get_stats())

    # 创建事件用于同步线程
    start_event = threading.Event()
    end_events = [threading.Event() for _ in range(100)]

    def worker(worker_id, end_event):
        """工作线程函数，故意长时间持有连接"""
        print(f"Worker {worker_id} 等待开始...")
        start_event.wait()  # 等待所有线程就绪

        try:
            print(f"Worker {worker_id} 尝试获取连接...")
            with ConnectionManager.get_connection() as conn:
               print(f"Worker {worker_id} 获取到连接 | 当前池状态:", ConnectionManager._pool.get_stats())

               # 模拟长时间操作
               time.sleep(3)
               flights = list(conn.list_flights())
               print(f"Worker {worker_id} 完成操作，找到 {len(flights)} flights")

        except Exception as e:
            print(f"Worker {worker_id} 出错: {e}")
        finally:
            end_event.set()  # 标记完成

    with ThreadPoolExecutor(max_workers=100) as executor:
        futures = []
        for i in range(100):
            future = executor.submit(worker, i, end_events[i])
            futures.append(future)

        # 给线程一点时间启动
        time.sleep(1)
        print("\n所有线程准备就绪，即将同时请求连接...\n")
        start_event.set()  # 同时释放所有线程

        # 等待所有线程完成
        for event in end_events:
            event.wait()

    print("\n=== 测试完成 ===")
    print("最终连接池状态:", ConnectionManager._pool.get_stats())

    ConnectionManager._pool.close_all()  # 清理连接池


# if __name__ == "__main__":
#     stress_test_connection_pool()


import pyarrow as pa
import pyarrow.ipc as ipc
import os
import numpy as np
import netCDF4

from dfwriter.abstract_writer import BaseWriter
from utils.logger_utils import get_logger
logger = get_logger(__name__)


class NCWriter(BaseWriter):
    """
    netcdf file writer implementing the BaseWriter interface.
    """
    def write(self, table: pa.Table, output_path: str):
        """
        将 Arrow Table 写回 NetCDF 文件。大文件非常慢。
        支持变量属性、全局属性、缺测值、原始dtype和shape的还原。
        分批写入，避免内存溢出。自动处理 datetime64[ns] 类型为 float64。
        支持还原压缩参数，保证文件大小与原始一致。
        """
        import gc
        try:
            meta = table.schema.metadata or {}

            def _meta_eval(val, default):
                if isinstance(val, bytes):
                    return eval(val.decode())
                elif isinstance(val, str):
                    return eval(val)
                else:
                    return default

            def get_meta(meta, key, default):
                if key in meta:
                    return meta[key]
                if isinstance(key, str) and key.encode() in meta:
                    return meta[key.encode()]
                if isinstance(key, bytes) and key.decode() in meta:
                    return meta[key.decode()]
                return default

            shapes = _meta_eval(get_meta(meta, 'shapes', '[]'), [])
            dtypes = _meta_eval(get_meta(meta, 'dtypes', '[]'), [])
            var_names = _meta_eval(get_meta(meta, 'var_names', '[]'), [])
            var_attrs = _meta_eval(get_meta(meta, 'var_attrs', '{}'), {})
            fill_values = _meta_eval(get_meta(meta, 'fill_values', '{}'), {})
            global_attrs = _meta_eval(get_meta(meta, 'global_attrs', '{}'), {})
            orig_lengths = _meta_eval(get_meta(meta, 'orig_lengths', '[]'), [])
            var_dims = _meta_eval(get_meta(meta, 'var_dims', '{}'), {})
            var_compress = _meta_eval(get_meta(meta, 'var_compress', '{}'), {})

            logger.info(f"写入 NetCDF 文件: {output_path}")
            logger.info(f"变量名: {var_names}")

            if not (len(var_names) == len(shapes) == len(dtypes) == len(orig_lengths)):
                raise ValueError(
                    f"元数据长度不一致: var_names({len(var_names)}), shapes({len(shapes)}), dtypes({len(dtypes)}), orig_lengths({len(orig_lengths)})"
                )

            # 动态估算 batch_size
            def estimate_row_bytes(shapes, dtypes):
                total = 0
                for shape, dtype in zip(shapes, dtypes):
                    n = int(np.prod(shape[1:])) if len(shape) > 1 else 1
                    total += np.dtype(dtype).itemsize * n
                return total

            try:
                import psutil
                avail_mem = psutil.virtual_memory().available
                target_mem = avail_mem // 4
            except Exception:
                target_mem = 512 * 1024 * 1024  # 默认512MB

            row_bytes = estimate_row_bytes(shapes, dtypes)
            batch_size = max(1000, min(100000, target_mem // max(row_bytes, 1)))
            logger.info(f"动态设置 batch_size={batch_size}，单行约{row_bytes}字节，可用内存目标{target_mem}字节")

            with netCDF4.Dataset(output_path, 'w') as ds:
                # 1. 创建所有维度
                for i, name in enumerate(var_names):
                    dims = var_dims.get(name, [f"{name}_dim{j}" for j in range(len(shapes[i]))])
                    shape = shapes[i]
                    for dim_name, dim_len in zip(dims, shape):
                        if dim_name not in ds.dimensions:
                            ds.createDimension(dim_name, dim_len)
                # 2. 创建所有变量
                nc_vars = []
                dtype_map = []
                for i, name in enumerate(var_names):
                    shape = shapes[i]
                    dtype = dtypes[i]
                    fill_value = fill_values.get(name, None)
                    dims = var_dims.get(name, [f"{name}_dim{j}" for j in range(len(shape))])
                    np_dtype = np.dtype(dtype)
                    attrs = var_attrs.get(name, {})
                    # 处理 datetime64[ns] 类型
                    if np.issubdtype(np_dtype, np.datetime64):
                        logger.warning(f"{name}: datetime64[ns] 不被 netCDF4 支持，自动转为 float64（单位：天）")
                        np_dtype = np.float64
                        attrs['units'] = attrs.get('units', 'days since 1970-01-01')
                    dtype_map.append(np_dtype)
                    # 还原压缩参数
                    compress_info = var_compress.get(name, {})
                    create_kwargs = {}
                    for k in ['zlib', 'complevel', 'shuffle', 'chunksizes']:
                        if k in compress_info and compress_info[k] is not None:
                            create_kwargs[k] = compress_info[k]
                    if fill_value is not None:
                        var = ds.createVariable(name, np_dtype, dims, fill_value=fill_value, **create_kwargs)
                    else:
                        var = ds.createVariable(name, np_dtype, dims, **create_kwargs)
                    nc_vars.append(var)
                    var_attrs[name] = attrs  # 更新属性，后续写属性用
                # 3. 分批写入数据
                write_offsets = [0 for _ in var_names]
                batch_count = 0 # 记录批次数
                logger.info(f"开始分批写入数据")
                for batch_idx, batch in enumerate(table.to_batches(batch_size)):
                    logger.info(f"处理第 {batch_idx + 1} 批数据")
                    batch_count += 1
                    for i, arr in enumerate(batch.columns):
                        arr_np = arr.to_numpy(zero_copy_only=False)
                        orig_length = orig_lengths[i]
                        remain = orig_length - write_offsets[i]
                        shape_i = shapes[i]
                        np_dtype = dtype_map[i]
                        other_dim = int(np.prod(shape_i[1:])) if len(shape_i) > 1 else 1
                        max_write_rows = len(arr_np) // other_dim
                        write_len = min(remain, max_write_rows)
                        if write_len <= 0:
                            continue
                        arr_write = arr_np[:write_len * other_dim]
                        # 类型转换和特殊处理
                        if np.issubdtype(np_dtype, np.datetime64):
                            arr_write = arr_write.astype('datetime64[D]').astype('float64')
                        # 兼容 parse/sample 阶段已转 float 的 cftime.datetime
                        # 如果 units 属性是 days since 1970-01-01，直接写 float64
                        if np.issubdtype(np_dtype, np.integer) and fill_values.get(var_names[i], None) is not None:
                            arr_write = np.where(np.isnan(arr_write), fill_values[var_names[i]], arr_write)
                            arr_write = arr_write.astype(np_dtype)
                        else:
                            arr_write = arr_write.astype(np_dtype)
                        arr_write = arr_write.reshape((write_len,) + tuple(shape_i[1:]))
                        nc_vars[i][write_offsets[i]:write_offsets[i]+write_len, ...] = arr_write
                        write_offsets[i] += write_len
                        del arr_write  # 只在定义后删除
                    del batch, arr_np
                    gc.collect()
                logger.info(f"共写入 {batch_count} 批数据")
                # 4. 写变量属性
                for i, name in enumerate(var_names):
                    attrs = var_attrs.get(name, {})
                    for k, v in attrs.items():
                        if k == "_FillValue":
                            continue
                        try:
                            nc_vars[i].setncattr(k, v)
                        except Exception:
                            logger.warning(f"变量 {name} 属性 {k}={v} 写入失败")
                # 5. 写全局属性
                for k, v in global_attrs.items():
                    try:
                        ds.setncattr(k, v)
                    except Exception:
                        logger.warning(f"全局属性 {k}={v} 写入失败")
            logger.info(f"写入 NetCDF 文件到 {output_path}")
        except Exception as e:
            logger.error(f"写入 NetCDF 文件失败: {e}")
            raise


from io import BytesIO

from parser.abstract_parser import BaseParser
import pyarrow as pa
import pyarrow.ipc as ipc
from utils.logger_utils import get_logger
logger = get_logger(__name__)

class DefaultParser(BaseParser):

    def parse(self, file_path: str) -> None:
        raise NotImplementedError("CSVParser.parse method is not implemented yet.")

    def sample(self, file_path: str) -> None:
        raise NotImplementedError("CSVParser.sample method is not implemented yet.")

    def count(self, file_path: str) -> int:
        raise NotImplementedError("CSVParser.count method is not implemented yet.")


from gmssl import sm2, func
import base64
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
import os

class SM2Utils:
    def __init__(self, private_key=None, public_key=None):
        """
        初始化 SM2 工具类
        :param private_key: 私钥（16 进制字符串）
        :param public_key: 公钥（16 进制字符串）
        """
        self.private_key = private_key
        self.public_key = public_key
        self.sm2_crypt = sm2.CryptSM2(public_key=public_key, private_key=private_key)

    @staticmethod
    def generate_key_pair():
        """
        生成 SM2 密钥对
        :return: 公钥和私钥（16 进制字符串）
        """
        private_key_hex = func.random_hex(64)  # 生成 64 位随机私钥
        sm2_crypt = sm2.CryptSM2(private_key=private_key_hex, public_key="")  # 临时初始化
        public_key_hex = sm2_crypt._kg(int(private_key_hex, 16), sm2_crypt.ecc_table['g'])  # 通过私钥生成公钥

        # 转换为 Base64 格式
        private_key_base64 = base64.b64encode(bytes.fromhex(private_key_hex)).decode("utf-8")
        public_key_base64 = base64.b64encode(bytes.fromhex(public_key_hex)).decode("utf-8")

        return public_key_base64, private_key_base64

    @staticmethod
    def generate_base64_key_pair():
        # 生成椭圆曲线密钥对
        private_key = ec.generate_private_key(ec.SECP256R1())
        public_key = private_key.public_key()

        # 提取私钥
        private_numbers = private_key.private_numbers()
        private_value = private_numbers.private_value

        # 提取公钥的 X 和 Y 坐标
        public_numbers = public_key.public_numbers()
        x = public_numbers.x
        y = public_numbers.y

        # 转换为 Base64 格式
        private_key_base64 = base64.b64encode(private_value.to_bytes(32, byteorder="big")).decode("utf-8")
        public_key_base64 = base64.b64encode(x.to_bytes(32, byteorder="big") + y.to_bytes(32, byteorder="big")).decode(
            "utf-8")

        return public_key_base64, private_key_base64

    def encrypt(self, data):
        """
        使用公钥加密数据
        :param data: 待加密数据（字节类型）
        :return: 加密后的字节数据
        """
        return self.sm2_crypt.encrypt(data)

    def decrypt(self, encrypted_data):
        """
        使用私钥解密数据
        :param encrypted_data: 加密后的字节数据
        :return: 解密后的字节数据
        """
        return self.sm2_crypt.decrypt(encrypted_data)

# 示例用法
if __name__ == "__main__":
    private_key = ec.generate_private_key(ec.SECP256R1())
    public_key = private_key.public_key()
    # 对方的公钥（示例中使用自己的公钥）
    peer_public_key = public_key

    # 生成共享密钥
    shared_key = private_key.exchange(ec.ECDH(), peer_public_key)

    derived_key = HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=None,
        info=b'handshake data',
    ).derive(shared_key)

    # # 待加密数据
    # data = b'{"key": "value"}'
    #
    # # 使用 AES 加密数据
    # iv = os.urandom(16)  # 生成随机 IV
    # cipher = Cipher(algorithms.AES(derived_key), modes.CFB(iv))
    # encryptor = cipher.encryptor()
    # encrypted_data = encryptor.update(data) + encryptor.finalize()
    #
    # # 解密数据
    # decryptor = cipher.decryptor()
    # decrypted_data = decryptor.update(encrypted_data) + decryptor.finalize()
    #
    # print("原始数据:", data)
    # print("加密数据:", encrypted_data)
    # print("解密数据:", decrypted_data)
    #
    # # 生成密钥对
    # public_key_base64_0, private_key_base64_0 = SM2Utils.generate_base64_key_pair()
    # print("Public Key:", public_key_base64_0)
    # print("Private Key:", private_key_base64_0)

    public_key_base64 = "MFkwEwYHKoZIzj0CAQYIKoEcz1UBgi0DQgAEtzATutfBFy08TH0_M28LfDuL_6t8_L8qGzIroYcWInQpBasGTmcwuhfPR74N3iEjXyrNomRaRK4v74RAkVzjaw=="
    private_key_base64 = "MIGTAgEAMBMGByqGSM49AgEGCCqBHM9VAYItBHkwdwIBAQQgBRVb_SZbnA5yizZRmlqNZzr4lZY7E3oiFQWhdaP6zU2gCgYIKoEcz1UBgi2hRANCAAS3MBO618EXLTxMfT8zbwt8O4v_q3z8vyobMiuhhxYidCkFqwZOZzC6F89Hvg3eISNfKs2iZFpEri_vhECRXONr"

    public_key_hex = base64.b64decode(public_key_base64).hex()
    private_key_hex = base64.b64decode(private_key_base64).hex()

    # 初始化 SM2 工具类
    sm2_utils = SM2Utils(private_key=private_key, public_key=public_key)

    # 待加密的 JSON 数据
    json_data = '{"key": "value"}'.encode("utf-8")

    # 加密
    encrypted_data = sm2_utils.encrypt(json_data)
    print("Encrypted (Base64):", base64.b64encode(encrypted_data).decode("utf-8"))

    # 解密
    decrypted_data = sm2_utils.decrypt(encrypted_data)
    print("Decrypted:", decrypted_data)
    #print("Decrypted:", decrypted_data.decode("utf-8"))

    # 验证加密解密结果
    assert json_data == decrypted_data, "SM2 加密解密测试失败"


import os

class FairdConfig:
    def __init__(self, config_file):
        self.config = {}
        self.load_config(config_file)

    def load_config(self, config_file):
        if not os.path.exists(config_file):
            raise FileNotFoundError(f"faird配置文件 {config_file} 不存在")
        with open(config_file, 'r', encoding='utf-8') as file:
            for line in file:
                line = line.strip()
                if line and not line.startswith('#') and not line.startswith('['):
                    key, value = line.split('=', 1)
                    self.config[key.strip()] = value.strip()

    def get(self, key, default=None):
        return self.config.get(key, default)

    @property
    def name(self):
        return self.get('host.name')

    @property
    def title(self):
        return self.get('host.title')

    @property
    def position(self):
        pos = self.get('host.position')
        return tuple(map(float, pos.split(','))) if pos else None

    @property
    def location(self):
        return self.get('host.location')

    @property
    def domain(self):
        return self.get('host.domain')

    @property
    def port(self):
        return int(self.get('host.port', 0))

    @property
    def external_domain(self):
        return self.get('host.external.domain')

    @property
    def external_port(self):
        return int(self.get('host.external.port', 0))

    @property
    def controld_address(self):
        return self.get('controld.address', '')

    @property
    def domain_id(self):
        return self.get('domain_id', '')

    @property
    def register_to_controld(self):
        return self.get('register_to_controld', 'false').lower()

    @property
    def public_key(self):
        return self.get('public_key')

    @property
    def log_path(self):
        return self.get('log.path')

    @property
    def metacat_url(self):
        return self.get('metacat_url')

    @property
    def metacat_token(self):
        return self.get('metacat_token')

    @property
    def access_mode(self):
        return self.get('access_mode', 'neo4j')

    @property
    def mongo_db_url(self):
        return self.get('mongo_db_url')

    @property
    def storage_local_path(self):
        return self.get('storage.local.path')

    @property
    def instrument_info(self):
        return self.get('instrument.info')

    @property
    def network_link_info(self):
        return self.get('network.link.info')

    @property
    def neo4j_url(self):
        return self.get('neo4j_url')

    @property
    def neo4j_user(self):
        return self.get('neo4j_user')

    @property
    def neo4j_password(self):
        return self.get('neo4j_password')

    @property
    def data_mongodb_host(self):
        return self.get('data.mongodb.host', '')

    @property
    def data_mongodb_port(self):
        return self.get('data.mongodb.port', '')

    @property
    def data_mongodb_database(self):
        return self.get('data.mongodb.database', '')

    @property
    def data_mongodb_collection(self):
        return self.get('data.mongodb.collection', '')

    @property
    def data_mongodb_username(self):
        return self.get('data.mongodb.username', '')

    @property
    def data_mongodb_password(self):
        return self.get('data.mongodb.password', '')

    @property
    def dataset_base_path_field(self):
        return self.get('dataset.base.path.field', '')

class FairdConfigManager:
    _config = None

    @classmethod
    def load_config(cls, config_file):
        cls._config = FairdConfig(config_file)

    @classmethod
    def get_config(cls):
        if cls._config is None:
            raise Exception("Faird Config not loaded")
        return cls._config


from tabulate import tabulate

def format_arrow_table(arrow_table, head_rows=5, tail_rows=5, first_cols=3, last_cols=3, display_all=False):
    if display_all:
        return arrow_table.to_pandas().to_string()
    all_columns = arrow_table.column_names
    total_columns = len(all_columns)
    total_rows = arrow_table.num_rows

    if total_columns <= (first_cols + last_cols):
        display_columns = all_columns
    else:
        display_columns = all_columns[:first_cols] + ['...'] + all_columns[-last_cols:]

    head_rows = min(head_rows, total_rows)
    tail_rows = min(tail_rows, total_rows)

    table_data = []
    head_data = arrow_table.slice(0, head_rows)

    for i in range(head_rows):
        row = []
        for col in display_columns:
            if col == '...':
                row.append('...')
            else:
                col_index = all_columns.index(col)
                row.append(str(head_data.column(col_index)[i].as_py()))
        table_data.append(row)

    if total_rows > (head_rows + tail_rows):
        table_data.append(['...' for _ in display_columns])

    if total_rows > head_rows:
        tail_data = arrow_table.slice(total_rows - tail_rows, tail_rows)
        for i in range(tail_rows):
            row = []
            for col in display_columns:
                if col == '...':
                    row.append('...')
                else:
                    col_index = all_columns.index(col)
                    row.append(str(tail_data.column(col_index)[i].as_py()))
            table_data.append(row)

    table_str = tabulate(table_data, headers=display_columns, tablefmt="plain")
    table_str += f"\n\n[{total_rows} rows x {total_columns} columns]"
    return table_str



from io import BytesIO

from dfwriter.abstract_writer import BaseWriter
import pyarrow as pa
import pyarrow.ipc as ipc
from utils.logger_utils import get_logger
logger = get_logger(__name__)


class DefaultWriter(BaseWriter):
    """
    Default file writer implementing the BaseWriter interface.
    """
    def write(self, table: None, output_path: str) -> None:
        try:
            if isinstance(output_path, BytesIO):
                # 使用 BufferOutputStream 写入字节流
                sink = pa.BufferOutputStream()
                with ipc.new_file(sink, table.schema) as writer:
                    writer.write_table(table)
                # 将数据写入 BytesIO
                output_path.write(sink.getvalue().to_pybytes())
            else:
                logger.info(f"即将写入 arrow 文件: {output_path}")
                with pa.OSFile(output_path, 'wb') as sink:
                    with ipc.new_file(sink, table.schema) as writer:
                        writer.write_table(table)
                logger.info(f"成功写入 arrow 文件: {output_path}")
        except Exception as e:
            logger.error(f"写入 arrow 文件时出错: {e}")
            raise


import logging

def get_logger(name):
    """
    获取主日志 logger（写入 faird.log）
    """
    logger = logging.getLogger(f"{name}_faird")
    if not logger.handlers:
        logger.setLevel(logging.INFO)

        # 主日志文件 handler
        log_handler = logging.FileHandler("faird.log")
        log_handler.setLevel(logging.INFO)
        log_handler.setFormatter(logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s [%(filename)s:%(lineno)d] - %(message)s"
        ))

        # 控制台日志 handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s [%(filename)s:%(lineno)d] - %(message)s"
        ))

        logger.addHandler(log_handler)
        logger.addHandler(console_handler)
    return logger

def get_access_logger(name):
    """
    获取访问日志 logger（写入 access.log）
    """
    logger = logging.getLogger(f"{name}_access")
    if not logger.handlers:
        logger.setLevel(logging.INFO)

        # 访问日志文件 handler
        access_handler = logging.FileHandler("access.log")
        access_handler.setLevel(logging.INFO)
        access_handler.setFormatter(logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s [%(filename)s:%(lineno)d] - %(message)s"
        ))

        # 控制台日志 handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s [%(filename)s:%(lineno)d] - %(message)s"
        ))

        logger.addHandler(access_handler)
        logger.addHandler(console_handler)
    return logger


import os
import numpy as np
import pyarrow as pa
import pyarrow.ipc as ipc
import tifffile
from parser.abstract_parser import BaseParser
from utils.logger_utils import get_logger
logger = get_logger(__name__)

class TIFParser(BaseParser):
    def parse(self, file_path: str) -> pa.Table:
        """
        读取 TIFF 文件为 Arrow Table，保留原始 dtype、shape、波段信息。
        支持多页、多波段。先写入.arrow缓存，再读取返回。
        """
        try:
            DEFAULT_ARROW_CACHE_PATH = os.path.expanduser("~/.cache/faird/dataframe/tif/")
            os.makedirs(DEFAULT_ARROW_CACHE_PATH, exist_ok=True)
            arrow_file_name = os.path.basename(file_path).rsplit(".", 1)[0] + ".arrow"
            arrow_file_path = os.path.join(DEFAULT_ARROW_CACHE_PATH, arrow_file_name)
            if os.path.exists(arrow_file_path):
                logger.info(f"检测到缓存文件，直接从 {arrow_file_path} 读取 Arrow Table。")
                try:
                    with pa.memory_map(arrow_file_path, "r") as source:
                        return ipc.open_file(source).read_all()
                except Exception as e:
                    logger.warning(f"读取缓存 .arrow 文件失败，将重新解析TIFF: {e}")

            logger.info(f"开始解析 TIFF 文件: {file_path}")

            pa_arrays_raw = []
            band_names = []
            orig_lengths = []
            pa_types = []
            page_shapes = []  # <--- 新增：存储每个原始页面的完整shape
            dtypes = []

            # --- 修复核心：使用 TiffFile 逐页读取并检查元数据标签 ---
            with tifffile.TiffFile(file_path) as tif:
                for idx, page in enumerate(tif.pages):
                    try:
                        img = page.asarray()
                        dtype = img.dtype
                        pa_type = pa.from_numpy_dtype(dtype)

                        # 记录该页的完整、原始的形状，这是还原的关键
                        page_shapes.append(img.shape)
                        dtypes.append(str(dtype))

                        # SamplesPerPixel: 每像素样本数，即波段数
                        samples_per_pixel = page.tags.get('SamplesPerPixel', 1).value

                        if samples_per_pixel == 1:  # 单波段/灰度图
                            arr = img.flatten()
                            pa_arrays_raw.append(pa.array(arr, type=pa_type))
                            pa_types.append(pa_type)
                            band_names.append(f'page{idx + 1}_band1')
                            orig_lengths.append(arr.size)
                        else:  # 多波段
                            # PlanarConfiguration: 1=Chunky(H,W,B), 2=Planar(B,H,W)
                            planar_config = page.tags.get('PlanarConfiguration', 1).value

                            # Chunky/Interleaved format (H, W, Bands)
                            if planar_config == 1:
                                for b in range(samples_per_pixel):
                                    arr = img[:, :, b].flatten()
                                    pa_arrays_raw.append(pa.array(arr, type=pa_type))
                                    pa_types.append(pa_type)
                                    band_names.append(f'page{idx + 1}_band{b + 1}')
                                    orig_lengths.append(arr.size)
                            # Planar format (Bands, H, W)
                            elif planar_config == 2:
                                for b in range(samples_per_pixel):
                                    arr = img[b, :, :].flatten()
                                    pa_arrays_raw.append(pa.array(arr, type=pa_type))
                                    pa_types.append(pa_type)
                                    band_names.append(f'page{idx + 1}_band{b + 1}')
                                    orig_lengths.append(arr.size)
                    except Exception as e:
                        logger.error(f"处理第{idx + 1}页时异常: {e}")
                        raise

            # --- 之后的 padding 逻辑保持不变 ---
            max_len = max(len(arr) for arr in pa_arrays_raw) if pa_arrays_raw else 0

            pa_arrays = []
            for arr, typ in zip(pa_arrays_raw, pa_types):
                try:
                    if len(arr) < max_len:
                        if pa.types.is_floating(typ):
                            padded = np.full(max_len, np.nan, dtype=typ.to_pandas_dtype())
                        else:
                            padded = np.zeros(max_len, dtype=typ.to_pandas_dtype())
                        padded[:len(arr)] = arr.to_numpy()
                        pa_arrays.append(pa.array(padded, type=typ))
                    else:
                        pa_arrays.append(arr)
                except Exception as e:
                    logger.error(f"补齐列时异常: {e}")
                    raise

            meta = {
                "page_shapes": str(page_shapes),  # 使用新的 page_shapes
                "dtypes": str(dtypes),
                "orig_lengths": str(orig_lengths),
                "band_names": str(band_names)
            }
            try:
                schema = pa.schema([pa.field(n, t) for n, t in zip(band_names, pa_types)]).with_metadata(
                    {k: str(v).encode() for k, v in meta.items()}
                )
                table = pa.table(pa_arrays, schema=schema)
                logger.info(
                    f"TIFF 解析完成，列数: {len(table.column_names)}，每列长度: {max_len}，写入缓存 {arrow_file_path}")
                with ipc.new_file(arrow_file_path, schema) as writer:
                    writer.write_table(table)
                # 再从.arrow读取返回
                with pa.memory_map(arrow_file_path, "r") as source:
                    return ipc.open_file(source).read_all()
            except Exception as e:
                logger.error(f"写入或读取 Arrow 缓存异常: {e}")
                raise
        except Exception as e:
            logger.error(f"TIFF 解析失败: {e}")
            raise

    def sample(self, file_path):
        """
        从 TIFF 文件中采样数据，返回 Arrow Table。
        读取第一页（或第一波段）前max_rows个像素，自动补齐为相同长度，并添加schema的metadata。
        max_rows: 生成的Arrow Table的行数，默认20（写在方法内部）
        """
        try:
            max_rows = 20
            logger.info(f"开始采样 TIFF 文件: {file_path}，采样行数: {max_rows}")
            try:
                with tifffile.TiffFile(file_path) as tif:
                    if len(tif.pages) == 0:
                        logger.error("TIFF 文件无有效页")
                        raise ValueError("TIFF 文件无有效页")
                    img = tif.pages[0].asarray()
            except Exception as e:
                logger.error(f"TIFF 采样读取第一页异常: {e}")
                raise
            logger.info(f"第一页 shape: {img.shape}, dtype: {img.dtype}")
            arrays = []
            names = []
            shapes = []
            dtypes = []
            orig_lengths = []
            try:
                if img.ndim == 2:
                    arr = img.flatten().astype(np.float64)
                    arrays.append(arr[:max_rows])
                    names.append('page1_band1')
                    shapes.append(img.shape)
                    dtypes.append(str(img.dtype))
                    orig_lengths.append(min(max_rows, arr.size))
                    logger.info("采样二维影像，波段数: 1")
                elif img.ndim == 3:
                    # (B, H, W)
                    if img.shape[0] in [1, 3, 4] and img.shape[0] < img.shape[1] and img.shape[0] < img.shape[2]:
                        logger.info(f"采样三维影像，按(B, H, W)模式，波段数: {img.shape[0]}")
                        for b in range(img.shape[0]):
                            arr = img[b, :, :].flatten().astype(np.float64)
                            arrays.append(arr[:max_rows])
                            names.append(f'page1_band{b+1}')
                            shapes.append(img[b, :, :].shape)
                            dtypes.append(str(img.dtype))
                            orig_lengths.append(min(max_rows, arr.size))
                    # (H, W, B)
                    elif img.shape[2] in [1, 3, 4] and img.shape[2] < img.shape[0] and img.shape[2] < img.shape[1]:
                        logger.info(f"采样三维影像，按(H, W, B)模式，波段数: {img.shape[2]}")
                        for b in range(img.shape[2]):
                            arr = img[:, :, b].flatten().astype(np.float64)
                            arrays.append(arr[:max_rows])
                            names.append(f'page1_band{b+1}')
                            shapes.append(img[:, :, b].shape)
                            dtypes.append(str(img.dtype))
                            orig_lengths.append(min(max_rows, arr.size))
                    else:
                        logger.info("采样三维影像，未知排列，直接flatten")
                        arr = img.flatten().astype(np.float64)
                        arrays.append(arr[:max_rows])
                        names.append('page1_flatten')
                        shapes.append(img.shape)
                        dtypes.append(str(img.dtype))
                        orig_lengths.append(min(max_rows, arr.size))
                else:
                    logger.info("采样高维影像，直接flatten")
                    arr = img.flatten().astype(np.float64)
                    arrays.append(arr[:max_rows])
                    names.append('page1_flatten')
                    shapes.append(img.shape)
                    dtypes.append(str(img.dtype))
                    orig_lengths.append(min(max_rows, arr.size))
            except Exception as e:
                logger.error(f"采样数据处理异常: {e}")
                raise
            # 补齐
            pa_arrays = []
            try:
                for arr in arrays:
                    if len(arr) < max_rows:
                        padded = np.full(max_rows, np.nan, dtype=np.float64)
                        padded[:len(arr)] = arr
                        pa_arrays.append(pa.array(padded))
                    else:
                        pa_arrays.append(pa.array(arr))
            except Exception as e:
                logger.error(f"采样数据补齐异常: {e}")
                raise
            # 构造schema并添加metadata
            try:
                schema = pa.schema([pa.field(n, pa.float64()) for n in names])
                meta = {
                    "shapes": str(shapes),
                    "dtypes": str(dtypes),
                    "orig_lengths": str(orig_lengths),
                    "file_type": "TIFF",
                    "sample": "True"
                }
                schema = schema.with_metadata({k: str(v).encode() for k, v in meta.items()})
                table = pa.table(pa_arrays, schema=schema)
                logger.info(f"采样完成，生成 Arrow Table，列: {names}，每列采样长度: {max_rows}，orig_lengths: {orig_lengths}")
                return table
            except Exception as e:
                logger.error(f"采样 Arrow Table 构造异常: {e}")
                raise
        except Exception as e:
            logger.error(f"采样 TIFF 文件失败: {e}")
            raise

    def count(self, file_path: str) -> int:
        """
        返回解析后 Arrow Table 的总行数（即所有页/波段像素数的最大值）。
        """
        try:
            images = tifffile.imread(file_path)
            # 统一成列表
            if images.ndim == 2:
                images = [images]
            elif images.ndim == 3:
                if images.shape[0] in [1, 3, 4] and images.shape[0] < images.shape[1] and images.shape[0] < images.shape[2]:
                    images = [images]
                elif images.shape[2] in [1, 3, 4] and images.shape[2] < images.shape[0] and images.shape[2] < images.shape[1]:
                    images = [images]
                else:
                    images = [img for img in images]
            elif images.ndim == 4:
                images = [img for img in images]
            else:
                images = [images]
            # 统计每个波段/页的像素数
            max_len = max(img.size for img in images)
            return int(max_len)
        except Exception as e:
            logger.error(f"统计 TIFF 文件 Arrow Table 行数失败: {e}")
            raise
        
    # def meta_to_json(self, meta: dict):
    #     """
    #     只返回 shape 和 dtype，变量为列，属性为行，适合前端表格展示。波段很少的很难看，只有一列，暂时不提供吧
    #     """
    #     import ast
    #     def safe_eval(val, default):
    #         try:
    #             return ast.literal_eval(val)
    #         except Exception:
    #             return default

    #     shapes = safe_eval(meta.get('shapes', '[]'), [])
    #     dtypes = safe_eval(meta.get('dtypes', '[]'), [])
    #     band_names = safe_eval(meta.get('band_names', '[]'), [])
    #     # 如果没有band_names就用 band1, band2...
    #     if not band_names:
    #         band_names = [f"band{i+1}" for i in range(len(shapes))]

    #     data = {}
    #     for i, v in enumerate(band_names):
    #         data[v] = {
    #             "shape": shapes[i] if i < len(shapes) else "",
    #             "dtype": dtypes[i] if i < len(dtypes) else ""
    #         }

    #     row_order = ["shape", "dtype"]

    #     result = {
    #         "columns": list(data.keys()),
    #         "rows": [
    #             {
    #                 "attribute": row,
    #                 **{col: data[col].get(row, "") for col in data}
    #             }
    #             for row in row_order
    #         ]
    #     }
    #     return result


from abc import ABC, abstractmethod
import pyarrow as pa

class BaseParser(ABC):
    """
    Abstract base class defining the interface for all parsers.
    """

    @abstractmethod
    def parse(self, file_path: str) -> pa.Table:
        """
        Parse the input file into a pyarrow.Table.

        Args:
            file_path (str): Path to the input file.
        Returns:
            pa.Table: A pyarrow Table object.
        """
        pass
    
    @abstractmethod
    def sample(self, file_path:str) -> pa.Table:
        """
        Sample the input file and return a pyarrow.Table.
        This method can be overridden by subclasses to provide sampling functionality.
        
        Args:
            file_path (str): Path to the input file.
        
        Returns:
            pa.Table: A pyarrow Table object containing sampled data.
        """
        pass
    
    @abstractmethod
    def count(self, file_path: str) -> int:
        """
        返回解析后 Arrow Table 的总行数。

        Args:
            file_path (str): 输入文件路径。

        Returns:
            int: Arrow Table 的总行数。
        """
        pass


from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
import base64

class RSAUtil:
    @staticmethod
    def generate_key_pair():
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        public_key = private_key.public_key()
        return public_key, private_key

    @staticmethod
    def save_private_key(private_key, path):
        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )
        with open(path, "wb") as f:
            f.write(pem)

    @staticmethod
    def save_public_key(public_key, path):
        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        with open(path, "wb") as f:
            f.write(pem)

    @staticmethod
    def load_private_key(path):
        with open(path, "rb") as f:
            return serialization.load_pem_private_key(f.read(), password=None)

    @staticmethod
    def load_public_key(path):
        with open(path, "rb") as f:
            return serialization.load_pem_public_key(f.read())

    @staticmethod
    def encrypt(public_key, data: bytes) -> bytes:
        return public_key.encrypt(
            data,
            padding.PKCS1v15()
        )

    @staticmethod
    def decrypt(private_key, ciphertext: bytes) -> bytes:
        return private_key.decrypt(
            ciphertext,
            padding.PKCS1v15()
        )

    @staticmethod
    def sign(private_key, data: bytes) -> bytes:
        return private_key.sign(
            data,
            padding.PKCS1v15(),
            hashes.SHA256()
        )

    @staticmethod
    def verify(public_key, data: bytes, signature: bytes) -> bool:
        try:
            public_key.verify(
                signature,
                data,
                padding.PKCS1v15(),
                hashes.SHA256()
            )
            return True
        except Exception:
            return False

def test_java_rsa():
    # Java RSA 密钥对
    # 粘贴Java输出的密文Base64
    cipher_b64 = "aYii16AHpVH0YNhXtu6Q/r9I3bUmccH7hEVuaqmzUWvzvwqPuYeH8VtBz4XsDSmCii2GTi+4ZOfYEe/QfIaNccGMjwUM5we6H4HfkXYTaKnBllgRnh9/RtzgGB2oEHXMHkX3Sep+r0HFgqp7xC3r+a1hQuGrewt8/97WVVKVfFuVarWncDmrUe4GKCgJz8zcINEpBi4NKu2/qLGs3hwh9iymfj1QZAheXXQP+xw3BYVkNT6rq3HYA0Ux0QIslsWv13+ud4fFEbFftVODIoPp72JB7qiq4Kq8xZcifCiVxC69tPLcYv4p99WOLZ9KPe7ysUEMqTEMA4tfa8LsBqvlqA=="
    ciphertext = base64.b64decode(cipher_b64)

    # 粘贴Java RSA 密钥对
    pub_pem = b"""-----BEGIN PUBLIC KEY-----
    MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAhcnlQGtech+JuUkU3l5t
    oB6yVMGfU2eVcL8sOV9SJoQvjd6K9mnMFYXV84eNECPki8VFPtx9X1X8u+FdSfI6
    WFle4ZS6pEkouxTBo2Q7vZyMFFTne0jiUHUB7ffABopnYEDYjspNFXK0C0Yl7wUc
    o/eOM7wFTNClvE0nR3AS488COyJCv5kXxgg2bTrcRsPKbU7/I/s1iXQpo/Rksgc4
    6ui3/VxOZ3nK5lmOiBFiM+Hd8RPqZnJWhRwnpTtDf8iknoYjofrY6h3mZIzf3gQD
    A33Xs0poZJA+/Nx2KdY2JScLEX8ZeuMkI5GjJuk4Zq4oUcEcCc2vGkRheOo3qYNK
    DwIDAQAB
    -----END PUBLIC KEY-----"""

    pri_pem = b"""-----BEGIN PRIVATE KEY-----
    MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQCFyeVAa15yH4m5
    SRTeXm2gHrJUwZ9TZ5Vwvyw5X1ImhC+N3or2acwVhdXzh40QI+SLxUU+3H1fVfy7
    4V1J8jpYWV7hlLqkSSi7FMGjZDu9nIwUVOd7SOJQdQHt98AGimdgQNiOyk0VcrQL
    RiXvBRyj944zvAVM0KW8TSdHcBLjzwI7IkK/mRfGCDZtOtxGw8ptTv8j+zWJdCmj
    9GSyBzjq6Lf9XE5necrmWY6IEWIz4d3xE+pmclaFHCelO0N/yKSehiOh+tjqHeZk
    jN/eBAMDfdezSmhkkD783HYp1jYlJwsRfxl64yQjkaMm6ThmrihRwRwJza8aRGF4
    6jepg0oPAgMBAAECggEAKHbLWSQAMTkbsQgLAUSDGIogOqVDGHk8SBjx+bo8QPz6
    2whDdErEWIAdmSXWwtpwPdvg2SXb33FE2thLHFngTqddHJIcryKXaJnbBR2EOhF2
    hpG1X2LIXDg/aKdlkzTaFlEWjn1eOg3yqXRCddvQcUEm57vGyJhUgDK/ngomFHMw
    TS9vNUXX63bb/b1kNs8JShW6D8nEamxXGOSg48kRI2z/UR8jTFngK91PMbuszRBy
    TZEAPuqyQMCxEecFNWHBixton0/6CHGW+8Fz0CRB9BBkqoKUYOvkYm5bxq76UPts
    U18vY9CybP/cuajp9tNUgg41/Yf+sLZC4+et2744cQKBgQC8QhBbqKY4UveNg8Lw
    m01TMuYaAou9pW8pU0X3g0ehkgddzgEdafuRffthcfxD941rEYXwSbeZyc42yG2i
    YBgowbOxyb7NdRyDRUbm0mZvDzl13Cw61J0Uj4UCNYTn9FQ7jcrHIKgVkKgRV+vj
    fB4cjSWRLhw2Dsn3JmN28J/tsQKBgQC17jl5+Onhnykg3HvQrJ2NcoNyasBLwBRZ
    y6uosoELJUTkwgDTAyxs06Rw3XNi3urAex8dYOCqzKnNkgDX7K5Vm/zu150DvBiO
    4Y79CPQ+oBMP8Hr0zLgbUYt5FMiGulKisC82qygZxNUncc+BuQJrVgSMpng5GUJo
    LSEbM33jvwKBgQCT2NpZ8bgg8d+izwDwzzbKCWY2tRsj5GHJtbF0MjC2TiXk4J11
    iT9dwvACVm+EvUzd8lQbIvnDGH8P/RZE+GzgzUcfcE6dssSiv2xwaHqz6+P3kONX
    jJSUiiYuVvO66uKgJrpC6n22+fVUMZV7AYXwbFU9kFXtim3I4Ogqh81B8QKBgFxs
    j8qFR9wVTYGiQUx7xPas7FZeIR7aj/wmWiiztpcA8gT9AgoxHIqU4n9sIUhjUwNC
    CUaVhs1+d+01dTQ8yzw2qMJ1yxCwqBE+SbQGrn56N/TrCKwArK9EOATlKBI56e04
    cGPng9HDiz4ktXSDltUMt/b0QzNOv8vI0X7OI/BlAoGANFRJSt+rzQnSst9536Ks
    3mHgklw6Te3kEJsGhZnUDjy8aNuWAj9Qd0Jp/rf1EAr3X9JJuAi6SO1uCULww/VB
    8g0E6Mq4GLx7sp6MhGuAwg8ZjeYkLypd4hj2MmYLWcvDKCNucz5ePzB1Xx20fcXq
    SORlppTr4W5KaARF/DkhsXI=
    -----END PRIVATE KEY-----"""

    pri = serialization.load_pem_private_key(pri_pem, password=None)

    # 解密
    dec = RSAUtil.decrypt(pri, ciphertext)
    print("解密明文:", dec.decode())

def test_generate_key_pair():
    # 生成密钥对
    pub, pri = RSAUtil.generate_key_pair()
    pri_pem = pri.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    pub_pem = pub.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    print("公钥PEM:\n", pub_pem.decode())
    print("私钥PEM:\n", pri_pem.decode())
    # 加密解密
    msg = b"hello from python"
    enc = RSAUtil.encrypt(pub, msg)
    print("密文Base64:", base64.b64encode(enc).decode())
    dec = RSAUtil.decrypt(pri, enc)
    print("解密明文:", dec.decode())
    # 签名验签
    sig = RSAUtil.sign(pri, msg)
    print("签名Base64:", base64.b64encode(sig).decode())
    print("验签:", RSAUtil.verify(pub, msg, sig))
    
    
    
if __name__ == "__main__":
    test_generate_key_pair()
    test_java_rsa()
    
    
    
    


import os
import numpy as np
import pyarrow as pa
import pyarrow.ipc as ipc
import xarray as xr
import dask
import netCDF4
import ast
import cftime
from parser.abstract_parser import BaseParser
from utils.logger_utils import get_logger
logger = get_logger(__name__)

def get_auto_chunk_size(var_shape, dtype=np.float64, target_mem_mb=10):
    dtype_size = np.dtype(dtype).itemsize
    if len(var_shape) == 0:
        return 1
    other_dim = int(np.prod(var_shape[1:])) if len(var_shape) > 1 else 1
    max_chunk = max(10, int((target_mem_mb * 1024 * 1024) // (other_dim * dtype_size)))
    return max_chunk

def is_large_variable(shape, dtype=np.float64, threshold_mb=50):
    dtype_size = np.dtype(dtype).itemsize
    size_mb = np.prod(shape) * dtype_size / 1024 / 1024
    return size_mb > threshold_mb

class NCParser(BaseParser):
    def parse(self, file_path: str) -> pa.Table:
        """
        用 xarray+dask 流式分块读取超大 NetCDF 文件，避免 OOM。
        并提取 dtype、压缩参数等元信息。
        兼容 _FillValue 和 missing_value 两种缺测值属性。
        保留原始缺测值（如 -9.96921e+36），不自动转为 np.nan。
        """
        DEFAULT_ARROW_CACHE_PATH = os.path.expanduser("~/.cache/faird/dataframe/csv/")
        # DEFAULT_ARROW_CACHE_PATH = os.path.join("D:/faird_cache/dataframe/nc/")
        os.makedirs(DEFAULT_ARROW_CACHE_PATH, exist_ok=True)
        arrow_file_name = os.path.basename(file_path).rsplit(".", 1)[0] + ".arrow"
        arrow_file_path = os.path.join(DEFAULT_ARROW_CACHE_PATH, arrow_file_name)

        file_size = os.path.getsize(file_path)
        logger.info(f"NetCDF 文件大小: {file_size} bytes")

        try:
            if os.path.exists(arrow_file_path):
                logger.info(f"检测到缓存文件，直接从 {arrow_file_path} 读取 Arrow Table。")
                with pa.memory_map(arrow_file_path, "r") as source:
                    return ipc.open_file(source).read_all()
        except Exception as e:
            logger.error(f"读取缓存 .arrow 文件失败: {e}")

        try:
            logger.info(f"开始用 xarray+dask 读取 NetCDF 文件: {file_path}")
            ds = xr.open_dataset(file_path, chunks={}, decode_cf=False)
            var_names = [v for v in ds.variables if ds[v].ndim > 0]
            shapes = [tuple(ds[v].shape) for v in var_names]
            dtypes = [str(ds[v].dtype) for v in var_names]
            var_attrs = {v: dict(ds[v].attrs) for v in var_names}
            def get_fill_value(attrs):
                for k in attrs:
                    if k.lower() in ['_fillvalue', 'missing_value']:
                        return attrs[k]
                return None
            fill_values = {v: get_fill_value(var_attrs[v]) for v in var_names}
            global_attrs = dict(ds.attrs)
            main_axes = [ds[v].dims[0] if ds[v].ndim > 0 else None for v in var_names]
            main_lens = [ds[v].shape[0] if ds[v].ndim > 0 else 1 for v in var_names]
            var_dims = {v: ds[v].dims for v in var_names}
            # 提取压缩参数
            var_compress = {}
            with netCDF4.Dataset(file_path) as nc:
                file_format = getattr(nc, 'file_format', 'unknown')
                for v in var_names:
                    var = nc.variables[v]
                    compress_info = {}
                    for attr in ['zlib', 'complevel', 'shuffle', 'chunksizes']:
                        if hasattr(var, attr):
                            compress_info[attr] = getattr(var, attr)
                    var_compress[v] = compress_info

            logger.info(f"变量列表: {var_names}")
            logger.info(f"变量 shapes: {shapes}")
            logger.info(f"变量 dtypes: {dtypes}")

            schema = pa.schema([pa.field(v, pa.from_numpy_dtype(ds[v].dtype)) for v in var_names])
            meta = {
                "shapes": str(shapes),
                "dtypes": str(dtypes),
                "var_names": str(var_names),
                "var_attrs": str(var_attrs),
                "fill_values": str(fill_values),
                "global_attrs": str(global_attrs),
                "orig_lengths": str(main_lens),
                "var_dims": str(var_dims),
                "file_type": file_format,
                "var_compress": str(var_compress)
            }
            schema = schema.with_metadata({k: str(v).encode() for k, v in meta.items()})

            large_vars = []
            small_vars = []
            for i, shape in enumerate(shapes):
                if is_large_variable(shape, dtype=np.float64, threshold_mb=50):
                    large_vars.append((i, var_names[i]))
                else:
                    small_vars.append((i, var_names[i]))

            max_chunks = []
            for i, shape in enumerate(shapes):
                chunk = get_auto_chunk_size(shape, dtype=np.float64, target_mem_mb=10)
                max_chunks.append(chunk)
            total_chunks = max([int(np.ceil(main_lens[i] / max_chunks[i])) for i in range(len(var_names))])

            logger.info(f"总分块数: {total_chunks}")

            with ipc.new_file(arrow_file_path, schema) as writer:
                for chunk_idx in range(total_chunks):
                    chunk_arrays = [None] * len(var_names)
                    chunk_lens = [0] * len(var_names)
                    logger.info(f"处理第 {chunk_idx+1}/{total_chunks} 块")
                    # 大变量串行
                    for i, v in large_vars:
                        safe_chunk_size = max_chunks[i]
                        main_dim = main_axes[i]
                        main_len = main_lens[i]
                        start = chunk_idx * safe_chunk_size
                        end = min(start + safe_chunk_size, main_len)
                        logger.debug(f"大变量 {v} 分块: start={start}, end={end}, shape={ds[v].shape}")
                        if start >= end:
                            arr_flat = np.array([], dtype=ds[v].dtype)
                        else:
                            darr = ds[v].isel({main_dim: slice(start, end)}).data
                            arr = darr.compute() if hasattr(darr, "compute") else np.array(darr)
                            arr_flat = np.array(arr).flatten()
                        chunk_arrays[i] = arr_flat
                        chunk_lens[i] = len(arr_flat)
                    # 小变量并行
                    batch_idxs = [i for i, _ in small_vars]
                    dask_chunks = []
                    safe_chunk_sizes = []
                    for i, v in small_vars:
                        safe_chunk_size = max_chunks[i]
                        main_dim = main_axes[i]
                        main_len = main_lens[i]
                        start = chunk_idx * safe_chunk_size
                        end = min(start + safe_chunk_size, main_len)
                        logger.debug(f"小变量 {v} 分块: start={start}, end={end}, shape={ds[v].shape}")
                        if start >= end:
                            dask_chunks.append(np.array([], dtype=ds[v].dtype))
                        else:
                            dask_chunks.append(ds[v].isel({main_dim: slice(start, end)}).data)
                        safe_chunk_sizes.append(safe_chunk_size)
                    computed_chunks = []
                    if dask_chunks:
                        try:
                            computed_chunks = dask.compute(*dask_chunks, scheduler="threads")
                        except Exception as e:
                            logger.warning(f"小变量并行失败，自动降级为串行: {e}")
                            computed_chunks = []
                            for chunk in dask_chunks:
                                computed_chunks.append(chunk.compute() if hasattr(chunk, "compute") else chunk)
                        for idx, arr in zip(batch_idxs, computed_chunks):
                            arr_flat = np.array(arr).flatten()
                            chunk_arrays[idx] = arr_flat
                            chunk_lens[idx] = len(arr_flat)
                    # 统一补齐到本次最大长度
                    max_len_this_chunk = max(chunk_lens) if chunk_lens else 0
                    for i, arr_flat in enumerate(chunk_arrays):
                        dtype = ds[var_names[i]].dtype
                        # cftime.datetime 类型兼容：转为自 1970-01-01 的天数
                        if arr_flat is not None and arr_flat.size > 0 and isinstance(arr_flat[0], cftime.datetime):
                            arr_flat = np.array([(x - cftime.DatetimeGregorian(1970, 1, 1)).days for x in arr_flat], dtype=np.float64)
                            chunk_arrays[i] = pa.array(arr_flat)
                            continue
                        if arr_flat is None:
                            padded = np.full(max_len_this_chunk, np.nan, dtype=dtype)
                            chunk_arrays[i] = pa.array(padded)
                        elif len(arr_flat) < max_len_this_chunk:
                            padded = np.full(max_len_this_chunk, np.nan, dtype=dtype)
                            padded[:len(arr_flat)] = arr_flat.astype(dtype)
                            chunk_arrays[i] = pa.array(padded)
                        else:
                            chunk_arrays[i] = pa.array(arr_flat.astype(dtype))
                    table = pa.table(chunk_arrays, names=var_names)
                    writer.write_table(table)
            ds.close()
            logger.info(f"Arrow Table 已写入缓存文件: {arrow_file_path}")
        except Exception as e:
            logger.error(f"解析 NetCDF 文件失败: {e}")
            raise

        try:
            logger.info(f"从 .arrow 文件 {arrow_file_path} 读取 Arrow Table。")
            with pa.memory_map(arrow_file_path, "r") as source:
                return ipc.open_file(source).read_all()
        except Exception as e:
            logger.error(f"读取 .arrow 文件失败: {e}")
            raise

    def sample(self, file_path: str) -> pa.Table:
        """
        从 NetCDF 文件中采样数据，返回 Arrow Table。
        默认每个变量只读取前10个主轴切片（如 time 维度的前10个）。
        更快的 NetCDF 采样方法：边采样边补齐，避免多余拷贝和类型推断。
        并为 schema 添加 metadata。
        兼容 _FillValue 和 missing_value 两种缺测值属性。
        保留原始缺测值（如 -9.96921e+36），不自动转为 np.nan。

        """
        try:
            # if file_path.startswith("/"):
            #     file_path = file_path[1:]
            ds = xr.open_dataset(file_path, decode_cf=False)
            var_names = [v for v in ds.variables if ds[v].ndim > 0]
            arrays = []
            field_types = []
            var_attrs = {v: dict(ds[v].attrs) for v in var_names}
            def get_fill_value(attrs):
                for k in attrs:
                    if k.lower() in ['_fillvalue', 'missing_value']:
                        return attrs[k]
                return None
            fill_values = {v: get_fill_value(var_attrs[v]) for v in var_names}
            max_len = 20
            for idx, v in enumerate(var_names):
                var = ds[v]
                if var.shape[0] > 10:
                    arr = var.isel({var.dims[0]: slice(0, 10)}).values
                else:
                    arr = var.values
                arr_flat = arr.flatten() if isinstance(arr, np.ndarray) else np.array(arr).flatten()
                # 类型推断和补齐
                if arr_flat.size > 0 and isinstance(arr_flat[0], cftime.datetime):
                    arr_flat = np.array([(x - cftime.DatetimeGregorian(1970, 1, 1)).days for x in arr_flat], dtype=np.float64)
                    typ = pa.float64()
                elif arr_flat.size > 0 and (isinstance(arr_flat[0], (bytes, np.bytes_, str))):
                    arr_flat = np.array([x.decode() if isinstance(x, (bytes, np.bytes_)) else str(x) for x in arr_flat], dtype=object)
                    typ = pa.string()
                else:
                    typ = pa.float64()
                field_types.append(typ)
                if typ == pa.string():
                    if len(arr_flat) >= max_len:
                        arrays.append(pa.array(arr_flat[:max_len], type=pa.string()))
                    else:
                        padded = np.full(max_len, "", dtype=object)
                        padded[:len(arr_flat)] = arr_flat
                        arrays.append(pa.array(padded, type=pa.string()))
                else:
                    fill_value = fill_values.get(v, np.nan)
                    if fill_value is None:
                        fill_value = np.nan
                    if len(arr_flat) >= max_len:
                        arrays.append(pa.array(arr_flat[:max_len].astype(np.float64), type=pa.float64()))
                    else:
                        padded = np.full(max_len, fill_value, dtype=np.float64)
                        padded[:len(arr_flat)] = arr_flat.astype(np.float64)
                        arrays.append(pa.array(padded, type=pa.float64()))
            schema = pa.schema([pa.field(v, t) for v, t in zip(var_names, field_types)])
            shapes = [tuple(ds[v].shape) for v in var_names]
            dtypes = [str(ds[v].dtype) for v in var_names]
            global_attrs = dict(ds.attrs)
            orig_lengths = [int(np.prod(ds[v].shape)) for v in var_names]
            var_dims = {v: ds[v].dims for v in var_names}
            meta = {
                "shapes": str(shapes),
                "dtypes": str(dtypes),
                "var_names": str(var_names),
                "var_attrs": str(var_attrs),
                "fill_values": str(fill_values),
                "global_attrs": str(global_attrs),
                "orig_lengths": str(orig_lengths),
                "var_dims": str(var_dims),
                "sample": "True"
            }
            schema = schema.with_metadata({k: str(v).encode() for k, v in meta.items()})
            table = pa.table(arrays, schema=schema)
            ds.close()
            return table
        except Exception as e:
            logger.error(f"采样 NetCDF 文件失败: {e}")
            raise

    def meta_to_json(self, meta: dict):
        """
        将 sample 方法生成的 meta 字典转为适合前端展示的 JSON 格式（变量为列，属性为行）。
        用法示例
        meta = {k.decode(): v.decode() for k, v in table.schema.metadata.items()}
        json_data = meta_to_json(meta)
        print(json_data)
        """
        def safe_eval(val, default):
            try:
                return ast.literal_eval(val)
            except Exception:
                return default

        shapes = safe_eval(meta.get('shapes', '[]'), [])
        dtypes = safe_eval(meta.get('dtypes', '[]'), [])
        var_names = safe_eval(meta.get('var_names', '[]'), [])
        var_attrs = safe_eval(meta.get('var_attrs', '{}'), {})
        fill_values = safe_eval(meta.get('fill_values', '{}'), {})
        var_dims = safe_eval(meta.get('var_dims', '{}'), {})
        orig_lengths = safe_eval(meta.get('orig_lengths', '[]'), [])
        global_attrs = safe_eval(meta.get('global_attrs', '{}'), {})

        # 组织每个变量的属性
        data = {}
        for i, v in enumerate(var_names):
            data[v] = {
                "shape": shapes[i] if i < len(shapes) else "",
                "dtype": dtypes[i] if i < len(dtypes) else "",
                "var_attrs": var_attrs.get(v, {}),
                "fill_value": fill_values.get(v, ""),
                "var_dims": var_dims.get(v, ""),
                "orig_length": orig_lengths[i] if i < len(orig_lengths) else ""
            }
        # 增加全局属性一列
        data["global_attrs"] = {
            "shape": "",
            "dtype": "",
            "var_attrs": "",
            "fill_value": "",
            "var_dims": "",
            "orig_length": "",
            "global_attrs": global_attrs
        }

        # 行顺序
        row_order = ["shape", "dtype", "var_attrs", "fill_value", "var_dims", "orig_length", "global_attrs"]

        # 转为前端友好的json
        result = {
            "columns": list(data.keys()),
            "rows": [
                {
                    "attribute": row,
                    **{col: data[col].get(row, "") for col in data}
                }
                for row in row_order
            ]
        }
        return result
    
    def count(self, file_path: str) -> int:
        """
        返回解析后 Arrow Table 的总行数（即主变量 shape 的元素数最大值）。
        新增：记录 Arrow Table 总行数，parse方法生成的Arrow Table总行数 = shape中最高维变量的所有元素数量（即所有主变量shape的乘积）！
        原理：parse方法通常会将所有主变量（shape大于1的变量）拉平成一维，每一行对应原始nc文件中所有主变量的一个元素（即所有维度的组合），而不是只按第一个维度分块。
        arrow_table_rows = max(np.prod(shape) for shape in shapes)
        """
        try:
            ds = xr.open_dataset(file_path, decode_cf=False)
            var_names = [v for v in ds.variables if ds[v].ndim > 0]
            shapes = [tuple(ds[v].shape) for v in var_names]
            if not shapes:
                return 0
            # Arrow Table 行数 = 所有主变量 shape 的元素总数的最大值
            total_rows = max(np.prod(shape) for shape in shapes)
            ds.close()
            return int(total_rows)
        except Exception as e:
            logger.error(f"统计 NetCDF 文件 Arrow Table 行数失败: {e}")
            raise


# faird/sdk/flight_client.py

import pyarrow.flight as flight


class FlightConnection:
    def __init__(self, host: str = "localhost", port: int = 8815):
        """
        初始化与远程 PyArrow Flight 服务的连接

        Args:
            host (str): 服务器地址
            port (int): 服务端口
        """
        self.host = host
        self.port = port
        self.client = flight.connect((host, port))

    def get_table(self, dataframe_id: str):
        """
        获取远程 Arrow Table

        Args:
            dataframe_id (str): 唯一标识符（文件名或 ID）

        Returns:
            pa.Table: 返回从远程读取的 Arrow 表格
        """
        ticket = flight.Ticket(dataframe_id.encode('utf-8'))
        reader = self.client.do_get(ticket)
        return reader.read_all()

    def write_table(self, table, dataframe_id: str):
        """
        将 Arrow Table 写入远程服务

        Args:
            table (pa.Table): 要上传的 Arrow 表格
            dataframe_id (str): 远程标识符
        """
        descriptor = flight.FlightDescriptor.for_path(dataframe_id)
        writer, _ = self.client.do_put(descriptor, table.schema)
        writer.write_table(table)
        writer.close()

    def get_schema(self, dataframe_id: str):
        """
        获取远程数据集的 schema

        Args:
            dataframe_id (str): 标识符

        Returns:
            Schema: Arrow schema 对象
        """
        descriptor = flight.FlightDescriptor.for_path(dataframe_id)
        info = self.client.get_flight_info(descriptor)
        return info.schema

    def list_datasets(self, criteria=None):
        """
        列出所有可用的数据集

        Args:
            criteria: 查询条件（可选）
        """
        flights = self.client.list_flights(criteria)
        for info in flights:
            logger.info(f"Dataset: {info.descriptor.path[0]}, Rows: {info.total_records}, Size: {info.total_bytes}")

    def do_action(self, action_type: str, body: bytes = b''):
        """
        执行一个远程操作（Action）

        Args:
            action_type (str): 操作类型（如 refresh_cache）
            body (bytes): 操作内容（可选）

        Returns:
            List[Any]: 操作结果列表
        """
        action = flight.Action(type=action_type, body=body)
        results = self.client.do_action(action)
        return [result.body.to_pybytes() for result in results]

    def close(self):
        """关闭连接"""
        self.client.close()



import math

import pyarrow as pa
import pyarrow.ipc as ipc
import os
import uuid
from parser.abstract_parser import BaseParser
from services.datasource.services import *
from core.config import FairdConfigManager
from utils.logger_utils import get_logger
logger = get_logger(__name__)

class DirParser(BaseParser):

    def sample_dir(self, file_path:str, dataset_name: str) -> pa.Table:
        data_source_service = None;
        if FairdConfigManager.get_config().access_mode == "interface":
            data_source_service = metacat_service.MetaCatService()
        elif FairdConfigManager.get_config().access_mode == "mongodb":
            data_source_service = metacat_mongo_service.MetaCatMongoService()
        elif FairdConfigManager.get_config().access_mode == "neo4j":
            data_source_service = metacat_neo4j_service.MetaCatNeo4jService()
        else:
            logger.error("Failed to load data source service.")

        files_data = []
        all_files = data_source_service.list_dataframes("", dataset_name, 1, 11)
        for file_dict in all_files:
            if file_dict["type"] == "dir":
                continue
            files_data.append({
                "name": file_dict["name"],
                "path": file_dict["path"],
                "suffix": file_dict["suffix"],
                "type": file_dict["type"],
                "size": file_dict["size"],
                "time": file_dict["time"] if file_dict.__contains__("time") else None,
                "blob": None
            })
        files_data = files_data[:10]

        # 定义 Arrow 表的 schema
        schema = pa.schema([
            ("name", pa.string()),
            ("path", pa.string()),
            ("suffix", pa.string()),
            ("type", pa.string()),
            ("size", pa.int64()),
            ("time", pa.string()),
            ("blob", pa.binary())
        ])
        table = pa.Table.from_pydict({key: [file[key] for file in files_data] for key in schema.names}, schema=schema)
        return table

    def parse_dir(self, file_path: str, dataset_name: str) -> pa.Table:
        # Ensure the cache directory exists
        DEFAULT_ARROW_CACHE_PATH = os.path.expanduser("~/.cache/faird/dataframe/dir/")
        os.makedirs(os.path.dirname(DEFAULT_ARROW_CACHE_PATH), exist_ok=True)

        arrow_file_name = str(uuid.uuid4()) + ".arrow"
        arrow_file_path = os.path.join(DEFAULT_ARROW_CACHE_PATH, arrow_file_name)

        data_source_service = None;
        if FairdConfigManager.get_config().access_mode == "interface":
            data_source_service = metacat_service.MetaCatService()
        elif FairdConfigManager.get_config().access_mode == "mongodb":
            data_source_service = metacat_mongo_service.MetaCatMongoService()
        elif FairdConfigManager.get_config().access_mode == "neo4j":
            data_source_service = metacat_neo4j_service.MetaCatNeo4jService()
        else:
            logger.error("Failed to load data source service.")

        #  分页
        total_length = data_source_service.get_dataframes_length(dataset_name)
        if total_length == 0:
            logger.info("total length is 0")
            return None
        limit = 10000
        total_pages = math.ceil(total_length / limit)
        all_files = []
        for page in range(1, total_pages + 1):
            dataframes = data_source_service.list_dataframes("", dataset_name, page=page, limit=limit)
            if dataframes:
                    all_files.extend(dataframes)

        files_data = []
        #all_files = data_source_service.list_dataframes("", dataset_name)
        for file_dict in all_files:
            if file_dict["type"] == "dir":
                continue
            files_data.append({
                "name": file_dict["name"],
                "path": file_dict["path"],
                "suffix": file_dict["suffix"],
                "type": file_dict["type"],
                "size": file_dict["size"],
                "time": file_dict["time"] if file_dict.__contains__("time") else None,
                "blob": None
            })

        # 定义 Arrow 表的 schema
        schema = pa.schema([
            ("name", pa.string()),
            ("path", pa.string()),
            ("suffix", pa.string()),
            ("type", pa.string()),
            ("size", pa.int64()),
            ("time", pa.string()),
            ("blob", pa.binary())
        ])
        table = pa.Table.from_pydict({key: [file[key] for file in files_data] for key in schema.names}, schema=schema)

        # Save the table as a .arrow file
        with ipc.new_file(arrow_file_path, table.schema) as writer:
            writer.write_table(table)

        # Load the .arrow file into a pyarrow Table using zero-copy
        with pa.memory_map(arrow_file_path, "r") as source:
            return ipc.open_file(source).read_all()

    def sample(self, file_path:str) -> pa.Table:
        return None

    def parse(self, file_path: str) -> pa.Table:
        return None

    def write(self, table: pa.Table, output_path: str):
        raise NotImplementedError("DirParser.write() 尚未实现：当前不支持写回 Dir 文件")

    def count(self, file_path: str) -> int:
        pass


import pyarrow as pa
import pyarrow.csv as csv
import pyarrow.ipc as ipc
import os

from parser.abstract_parser import BaseParser
from utils.logger_utils import get_logger
logger = get_logger(__name__)


class CSVParser(BaseParser):
    """
    CSV file parser implementing the BaseParser interface.
    """

    def sample(self, file_path:str) -> pa.Table:
        try:
            logger.info(f"即将读取 CSV 文件: {file_path}")
            table = csv.read_csv(
                file_path,
                read_options=csv.ReadOptions(block_size=1024 * 1024),  # 设置块大小,
                convert_options=csv.ConvertOptions(),
                parse_options=csv.ParseOptions()
            ).slice(0, 10) # 使用 slice 限制行数
            logger.info(f"成功读取 CSV 文件: {file_path}")
            return table
        except Exception as e:
            logger.error(f"读取 CSV 文件时出错: {e}")
            raise

    def parse(self, file_path: str) -> pa.Table:
        """
        Save the CSV file as a .arrow file and load it into memory as a pyarrow Table using zero-copy.

        Args:
            file_path (str): Path to the input CSV file.
        Returns:
            pa.Table: A pyarrow Table object.
        """

        # Ensure the cache directory exists
        DEFAULT_ARROW_CACHE_PATH = os.path.expanduser("~/.cache/faird/dataframe/csv/")
        os.makedirs(os.path.dirname(DEFAULT_ARROW_CACHE_PATH), exist_ok=True)

        # Extract the file name and append the .arrow suffix
        arrow_file_name = os.path.basename(file_path).rsplit(".", 1)[0] + ".arrow"
        arrow_file_path = os.path.join(DEFAULT_ARROW_CACHE_PATH, arrow_file_name)

        # Read the CSV file into a pyarrow Table
        try:
            logger.info(f"即将读取 CSV 文件: {file_path}")
            table = csv.read_csv(file_path)
            logger.info(f"成功读取 CSV 文件: {file_path}")
        except Exception as e:
            logger.info(f"读取 CSV 文件时出错: {e}")
            raise

        try:
            with ipc.new_file(arrow_file_path, table.schema) as writer:
                writer.write_table(table)
        except Exception as e:
            logger.info(f"保存 .arrow 文件时出错: {e}")
            raise

        try:
            with pa.memory_map(arrow_file_path, "r") as source:
                return ipc.open_file(source).read_all()
        except Exception as e:
            logger.info(f"加载 .arrow 文件时出错: {e}")
            raise

    def count(self, file_path: str) -> int:
        try:
            table = csv.read_csv(file_path)
            return table.num_rows
        except Exception as e:
            logger.error(f"统计 CSV 文件行数时出错: {e}")
            raise


from setuptools import setup, find_packages

setup(
    name="faird",
    version="1.1.10",
    description="A SDK for working with DataFrame",
    author="rdcn",
    author_email="rdcn@cnic.com",
    packages=find_packages(include=["core*", "parser*", "dfwriter*", "sdk*", "utils*"]),  # 自动发现包含的所有包
    install_requires=[
        "pyarrow==19.0.0",  # 指定依赖的 pyarrow 版本
        "tabulate",
        "pandas",
        "netCDF4",
        "rasterio",
        "requests",
        "duckdb",
        "pathlib"
    ],
    python_requires=">=3.9",
    entry_points={
        "console_scripts": [
            # 可选：定义命令行工具
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True
)


import pyarrow as pa
import pyarrow.ipc as ipc
import os
import numpy as np
import tifffile

from dfwriter.abstract_writer import BaseWriter
from utils.logger_utils import get_logger
logger = get_logger(__name__)


class TIFWriter(BaseWriter):
    """
    tif file writer implementing the BaseWriter interface.
    """

    def write(self, table: pa.Table, output_path: str):
        """
        将 Arrow Table 写入 TIFF 文件。
        通过读取元数据中的 'page_shapes' 来精确还原。
        """
        try:
            logger.info(f"开始写入 TIFF 文件: {output_path}")
            meta = table.schema.metadata or {}
            try:
                page_shapes = eval(meta.get(b'page_shapes', b'[]').decode())
                dtypes = eval(meta.get(b'dtypes', b'[]').decode())
                orig_lengths = eval(meta.get(b'orig_lengths', b'[]').decode())
            except Exception as e:
                logger.error(f"元数据解析异常: {e}")
                raise

            arrays = [col.to_numpy() for col in table.columns]
            images = []
            arr_idx = 0

            # --- 修复核心：基于 page_shapes 循环进行还原 ---
            for i, page_shape in enumerate(page_shapes):
                dtype = np.dtype(dtypes[i])

                # 情况一：原始页面是 2D (单波段)
                if len(page_shape) == 2:
                    if arr_idx >= len(arrays): raise IndexError("列数据不足以还原指定的页面形状")
                    valid_data = arrays[arr_idx][:orig_lengths[arr_idx]]
                    img = valid_data.reshape(page_shape).astype(dtype)
                    images.append(img)
                    arr_idx += 1

                # 情况二：原始页面是 3D (多波段)
                elif len(page_shape) == 3:
                    # 从已知的正确形状中确定波段数和排列方式
                    # Planar: (B, H, W) - 波段数在第0轴
                    if page_shape[0] < page_shape[1] and page_shape[0] < page_shape[2]:
                        num_bands = page_shape[0]
                        stack_axis = 0
                        h, w = page_shape[1], page_shape[2]
                    # Interleaved/Chunky: (H, W, B) - 波段数在最后1轴
                    else:
                        num_bands = page_shape[2]
                        stack_axis = -1
                        h, w = page_shape[0], page_shape[1]

                    if arr_idx + num_bands > len(arrays): raise IndexError("列数据不足以还原指定的多波段页面")

                    band_imgs = []
                    for _ in range(num_bands):
                        valid_data = arrays[arr_idx][:orig_lengths[arr_idx]]
                        band_img = valid_data.reshape((h, w)).astype(dtype)
                        band_imgs.append(band_img)
                        arr_idx += 1

                    # 按照正确的轴堆叠回多波段图像
                    img = np.stack(band_imgs, axis=stack_axis)
                    images.append(img)

                # 其他更高维度的情况
                else:
                    if arr_idx >= len(arrays): raise IndexError("列数据不足以还原指定的页面形状")
                    valid_data = arrays[arr_idx][:orig_lengths[arr_idx]]
                    img = valid_data.reshape(page_shape).astype(dtype)
                    images.append(img)
                    arr_idx += 1

            tifffile.imwrite(output_path, images if len(images) > 1 else images[0])
            logger.info(f"写入 TIFF 文件到 {output_path}，共 {len(images)} 页")
        except Exception as e:
            logger.error(f"写入 TIFF 文件失败: {e}")
            raise


import pyarrow as pa
import pyarrow.ipc as ipc
import os

from dfwriter.abstract_writer import BaseWriter
from utils.logger_utils import get_logger
logger = get_logger(__name__)


class CSVWriter(BaseWriter):
    """
    CSV file writer implementing the BaseWriter interface.
    """

    def write(self, table: pa.Table, output_path: str):
        try:
            logger.info(f"即将写入 CSV 文件: {output_path}")
            # 将 Arrow Table 转换为 Pandas DataFrame
            df = table.to_pandas()
            # 写入 CSV 文件
            df.to_csv(output_path, index=False)
            logger.info(f"成功写入 CSV 文件: {output_path}")
        except Exception as e:
            logger.error(f"写入 CSV 文件时出错: {e}")
            raise


from __future__ import annotations

from enum import Enum
from io import BytesIO
from typing import Optional, List
from urllib.parse import urlparse
import pyarrow as pa
import pyarrow.flight
import json
import socket
from utils.logger_utils import get_logger
logger = get_logger(__name__)
import requests
import zipfile
import os
from sdk.connection_pool import FlightConnectionPool, ConnectionManager
import datetime


class DacpClient:
    def __init__(self, url: str, principal: Optional[Principal] = None):
        self.__url = url
        self.__principal = principal
        self.__token = None
        self.__connection_id = None

    @staticmethod
    def connect(url: str, principal: Optional[Principal] = None) -> DacpClient:
        client = DacpClient(url, principal)
        logger.info(f"Connecting to {url} with principal {principal}...")
        parsed = urlparse(url)
        host = f"grpc://{parsed.hostname}:{parsed.port}"
        ConnectionManager.set_connection_pool(FlightConnectionPool(host, max_connections=20))

        # 构建ticket
        try:
            client_ip = socket.gethostbyname(socket.gethostname())
        except socket.gaierror:
            client_ip = "127.0.0.1"
        ticket = {
            'clientIp': client_ip
        }

        if principal:
            if principal.auth_type == AuthType.OAUTH:
                ticket.update({
                    'auth_type': principal.auth_type.value,
                    'type': principal.params.get('type'),
                    'username': principal.params.get('username'),
                    'password': principal.params.get('password')
                })
            elif principal.auth_type == AuthType.CONTROLD:
                ticket.update({
                    'auth_type': principal.auth_type.value,
                    'controld_domain_name': principal.params.get('controld_domain_name'),
                    'signature': principal.params.get('signature')
                })
            elif principal.auth_type == AuthType.ANONYMOUS:
                ticket.update({
                    'auth_type': principal.auth_type.value
                })

        # 发送连接请求
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("connect_server", json.dumps(ticket).encode('utf-8')))
            for res in results:
                res_json = json.loads(res.body.to_pybytes().decode('utf-8'))
                if res_json.get('errorMsg') != None:
                    logger.error(res_json.get('errorMsg'))
                    return None
                client.__token = res_json.get("token")
                client.__connection_id = res_json.get("connectionID")
            return client

    def list_datasets(self) -> List[str]:
        ticket = {
            'token': self.__token,
            'page': 1,
            'limit': 999999
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("list_datasets", json.dumps(ticket).encode('utf-8')))
            for res in results:
                res_json = json.loads(res.body.to_pybytes().decode('utf-8'))
                return res_json

    def get_dataset(self, dataset_name: str):
        ticket = {
            'token': self.__token,
            'dataset_name': dataset_name
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("get_dataset", json.dumps(ticket).encode('utf-8')))
            for res in results:
                res_json = json.loads(res.body.to_pybytes().decode('utf-8'))
                return res_json

    def list_dataframes(self, dataset_name: str) -> List[str]:
        ticket = {
            'token': self.__token,
            'username': self.__principal.params.get('username') if self.__principal and self.__principal.params else None,
            'dataset_name': dataset_name,
            'max_chunksize': 50000
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("list_dataframes", json.dumps(ticket).encode('utf-8')))
            dataframes = []
            for res in results:
                res_json = json.loads(res.body.to_pybytes().decode('utf-8'))
                dataframes.extend(res_json)
                print(f"Successfully fetch {len(dataframes)} dataframes")
            return dataframes

    def list_dataframes_stream(self, dataset_name: str, max_chunksize: Optional[int] = 50000):
        ticket = {
            'token': self.__token,
            'username': self.__principal.params.get('username') if self.__principal and self.__principal.params else None,
            'dataset_name': dataset_name,
            'max_chunksize': max_chunksize
        }
        with ConnectionManager.get_connection() as conn:
            reader = conn.do_action(pa.flight.Action("list_dataframes", json.dumps(ticket).encode('utf-8')))
            for chunk in reader:
                chunk_json = json.loads(chunk.body.to_pybytes().decode('utf-8'))
                print(f"Successfully fetch {len(chunk_json)} dataframes")
                yield chunk_json

    def list_user_auth_dataframes(self, username: str) -> List[str]:
        if username is None or username == "":
            logger.error("No username provided")
            return None
        ticket = {
            'username': username
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("list_user_auth_dataframes", json.dumps(ticket).encode('utf-8')))
            for res in results:
                res_json = json.loads(res.body.to_pybytes().decode('utf-8'))
                return res_json

    def check_permission(self, dataset_name: str, username: str) -> bool:
        ticket = {
            'dataset_name': dataset_name,
            'username': username
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("check_permission", json.dumps(ticket).encode('utf-8')))
            for res in results:
                res_str = res.body.to_pybytes().decode('utf-8')
                return res_str.lower() == 'true'

    def sample(self, dataframe_name: str):
        ticket = {
            'dataframe_name': dataframe_name,
            'connection_id': self.__connection_id
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("sample", json.dumps(ticket).encode('utf-8')))
            for res in results:
                return res.body.to_pybytes().decode('utf-8')

    def count(self, dataframe_name: str):
        ticket = {
            'dataframe_name': dataframe_name,
            'connection_id': self.__connection_id
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("count", json.dumps(ticket).encode('utf-8')))
            for res in results:
                return res.body.to_pybytes().decode('utf-8')

    def open(self, dataframe_name: str):
        from sdk.dataframe import DataFrame
        ticket = {
            'dataframe_name': dataframe_name,
            'connection_id': self.__connection_id
        }
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(pa.flight.Action("open", json.dumps(ticket).encode('utf-8')))
            return DataFrame(id=dataframe_name, connection_id=self.__connection_id)


    def get_ips(self) -> dict:
        public_ip = "0.0.0.0"
        private_ip = "127.0.0.1"

        # 获取公网IP
        try:
            response = requests.get('https://ifconfig.me/ip', timeout=5)
            response.raise_for_status()
            public_ip = response.text.strip()
        except requests.exceptions.RequestException as e:
            logger.warn(f"获取公网IP失败：{e}")

        # 获取内网IP
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            private_ip = s.getsockname()[0]
            s.close()
        except Exception as e:
            logger.warn(f"获取内网IP失败：{e}")

        return {
            "public_ip": public_ip,
            "private_ip": private_ip
        }

    # mount 挂载时候 ，验证并获取 token 有效期
    def token_verification(self, dataframe_name: str, token: str, username: Optional[str] = None):
        if not username:
            username = self.__principal.params.get('username')
        ticket = {
            'dataframe_name': dataframe_name,
            'token': token,
            'username': username,
            'connection_id': self.__connection_id
        }
        # 验证令牌是否有效
        with ConnectionManager.get_connection() as conn:
            results = conn.do_action(
                pa.flight.Action("token_verification", json.dumps(ticket).encode('utf-8')))
            for res in results:
                result_json = json.loads(res.body.to_pybytes().decode('utf-8'))
                if (result_json.get('exists')):
                    deadline_str = result_json.get('deadline')
                    token_info = {
                        "exists": True,
                        'deadline': deadline_str,
                        'usedTraffic': result_json.get('usedTraffic'),
                        'dataQuota': result_json.get('dataQuota')
                    }
                    return token_info
                else:
                    error_msg = "Token验证失败：不存在或已过期"
                    error_info = {
                        "exists": False,
                        "message": error_msg
                    }
                    return error_info
            return None

    def get_dataframe_stream(self, dataframe_name: str, token: str, max_chunksize: Optional[int] = 1024 * 1024 * 5
                             , username: Optional[str] = None , mount_timestamp: Optional[str] = None):
        if not token:
            raise ValueError("token is none")
        if not username :
            username = self.__principal.params.get('username')
        ips = self.get_ips()
        public_ip = ips['public_ip']
        private_ip = ips['private_ip']

        ticket = {
            'dataframe_name': dataframe_name,
            'token': token,
            'username': username,
            "public_ip": public_ip,
            "private_ip": private_ip,
            'max_chunksize': max_chunksize,
            'mount_timestamp': mount_timestamp,
            'connection_id': self.__connection_id
        }
        # 获取数据流
        with ConnectionManager.get_connection() as conn:
            stream_generator = conn.do_action(
                pa.flight.Action("get_dataframe_stream", json.dumps(ticket).encode('utf-8')))
            # 1. 尝试获取第一个数据块并检查其内容
            try:
                first_chunk = next(stream_generator)
                chunk_bytes = first_chunk.body.to_pybytes()
            except StopIteration:
                logger.error(f"数据获取失败: 没有数据可供获取")
                raise ValueError("数据获取失败: 没有数据可供获取")

            try:
                error_info = json.loads(chunk_bytes.decode('utf-8'))
                if 'status' in error_info and error_info['status'] == 'error' and error_info['code'] > 200:
                    logger.error(f"数据获取失败: {error_info['message']}")
                    raise ValueError(error_info)
            except (json.JSONDecodeError, UnicodeDecodeError):
                pass

            yield chunk_bytes
            for chunk in stream_generator:
                yield chunk.body.to_pybytes()

    def download_dataframe_zip(self, dacp_url: str, token: str, dest_folder_path:str, max_chunksize: Optional[int] = 1024 * 1024 * 5
                             , username: Optional[str] = None):

        dataframes = self.list_dataframes(dacp_url)
        file_name = urlparse(dacp_url).path.strip('/')

        zip_full_path = os.path.join(dest_folder_path, f"{file_name}.zip")

        if not os.path.exists(dest_folder_path):
            logger.error(f"Output directory not found: {dest_folder_path}")
            raise FileNotFoundError(f"Output directory not found: {dest_folder_path}")

        if os.path.exists(zip_full_path):
            logger.error(f"ZIP file already exists: {zip_full_path}")
            raise FileExistsError(f"ZIP file already exists: {zip_full_path}")
        try:
            with zipfile.ZipFile(zip_full_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
                for dataframe in dataframes:
                    file_path = os.path.relpath(dataframe['path'], dataframe['basePath']).replace(os.sep, '/').lstrip('/')

                    now = datetime.datetime.now()
                    zip_info = zipfile.ZipInfo(file_path, date_time=(now.year, now.month, now.day, now.hour, now.minute, now.second))
                    # 设置压缩方法
                    zip_info.compress_type = zipfile.ZIP_DEFLATED
                    total_size = 0
                    compress_size_start = zipf.fp.tell()
                    # 使用 zipf.open() 获取一个可写的流
                    with zipf.open(zip_info, 'w') as zip_file_stream:
                        for chunk in self.get_dataframe_stream(dataframe['dataframeName'], token=token, max_chunksize = max_chunksize, username=username):
                            zip_file_stream.write(chunk)
                            total_size += len(chunk)
                    zip_info.file_size = total_size
                    zip_info.compress_size = zipf.fp.tell() - compress_size_start

                    zipf.infolist().remove(zipf.getinfo(file_path))
                    zipf.infolist().append(zip_info)

                    logger.info(f"Added {file_path} (size: {total_size} bytes) to the ZIP archive.")
        except Exception as e:
            logger.error(f"Error creating ZIP archive: {e}", exc_info=True)
            if os.path.exists(zip_full_path):
                os.remove(zip_full_path)
            raise
        logger.info(f"Successfully created ZIP archive: {zip_full_path}")

class AuthType(Enum):
    OAUTH = "oauth"
    CONTROLD = "controld"
    ANONYMOUS = "anonymous"

class Principal:
    ANONYMOUS = None

    def __init__(self, auth_type: AuthType, **kwargs):
        self.auth_type = auth_type
        self.params = kwargs

    @staticmethod
    def oauth(type: str,  **kwargs) -> Principal:
        return Principal(AuthType.OAUTH, type=type, **kwargs)

    @staticmethod
    def controld(domain_name: str, signature: str, **kwargs) -> Principal:
        return Principal(AuthType.CONTROLD, controld_domain_name=domain_name, signature=signature, **kwargs)

    @staticmethod
    def anonymous() -> Principal:
        return Principal(AuthType.ANONYMOUS)

    def __repr__(self):
        return f"Principal(auth_type={self.auth_type}, params={self.params})"

Principal.ANONYMOUS = Principal.anonymous()





from abc import ABC, abstractmethod
import pyarrow as pa


class BaseWriter(ABC):
    """
    Abstract base class defining the interface for all parsers.
    """

    @abstractmethod
    def write(self, table: pa.Table, output_path: str):
        """
        Write the pyarrow.Table back to the original file format.

        Args:
            table (pa.Table): The Arrow Table to write.
            output_path (str): Output file path.
        """
        pass


import json
from typing import List, Optional, Dict, Any

import duckdb
import pandas
import pyarrow as pa
import pyarrow.compute as pc

from core.models.dataframe import DataFrame
from sdk.dacp_client import ConnectionManager
from utils.format_utils import format_arrow_table
import os


class DataFrame(DataFrame):

    def __init__(self, id: str, data: Optional[pa.Table] = None, actions: Optional[List[tuple]] = [], connection_id: Optional[str] = None):
        self.id = id
        self.data = data # 初始状态下 data 为空
        self.actions = actions # 用于记录操作的列表，延迟执行
        self.connection_id = connection_id

    def __str__(self) -> str:
        return self.to_string(head_rows=5, tail_rows=5, first_cols=3, last_cols=3)

    def __len__(self) -> int:
        return self.num_rows

    def __getitem__(self, index):
        if isinstance(index, int):  # 行选择
            if self.data is None:
                ticket = {
                    "dataframe": json.dumps(self, default=vars),
                    "row_index": index
                }
                with ConnectionManager.get_connection() as conn:
                    reader = conn.do_get(pa.flight.Ticket(json.dumps(ticket).encode('utf-8')))
                row_data = reader.read_all().to_pydict()
                #return {col: row_data[col][0] for col in row_data}
                row_data = {col: row_data[col][0] for col in row_data}

            #row_data = {col: self.data[col][index].as_py() for col in self.data.column_names}

                if "blob" in row_data and row_data["blob"] == b'':
                    ticket = {
                        "dataframe": json.dumps(self, default=vars),
                        "row_index": index,
                        "column_name": "blob"
                    }
                    with ConnectionManager.get_connection() as conn:
                        reader = conn.do_get(pa.flight.Ticket(json.dumps(ticket).encode('utf-8')))
                    # 返回流式读取生成器
                    def blob_stream():
                        for batch in reader:
                            for blob in batch.data.column(0):  # 遍历每个分块
                                yield blob.as_py()
                    row_data["blob"] = blob_stream()
                return row_data
        elif isinstance(index, str):  # 列选择
            if self.data is None:
                ticket = {
                    "dataframe": json.dumps(self, default=vars),
                    "column_name": index
                }
                with ConnectionManager.get_connection() as conn:
                    reader = conn.do_get(pa.flight.Ticket(json.dumps(ticket).encode('utf-8')))
                column_data = reader.read_all()
                return column_data.column(0).combine_chunks().to_pylist()  # 转换为列表
            return self.data[index].combine_chunks().to_pylist()
        else:
            raise TypeError("Index must be an integer (row) or string (column).")


    @property
    def schema(self):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars)
            }
            descriptor = pa.flight.FlightDescriptor.for_command(json.dumps(ticket))
            with ConnectionManager.get_connection() as conn:
                flight_info = conn.get_flight_info(descriptor)
            return flight_info.schema
        return self.data.schema

    @property
    def num_rows(self) -> int:
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars)
            }
            descriptor = pa.flight.FlightDescriptor.for_command(json.dumps(ticket))
            with ConnectionManager.get_connection() as conn:
                flight_info = conn.get_flight_info(descriptor)
            return flight_info.total_records
        return self.data.num_rows

    @property
    def num_cols(self):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars)
            }
            descriptor = pa.flight.FlightDescriptor.for_command(json.dumps(ticket))
            with ConnectionManager.get_connection() as conn:
                flight_info = conn.get_flight_info(descriptor)
            return len(flight_info.schema)
        return len(self.data.column_names)

    @property
    def shape(self):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars)
            }
            descriptor = pa.flight.FlightDescriptor.for_command(json.dumps(ticket))
            with ConnectionManager.get_connection() as conn:
                flight_info = conn.get_flight_info(descriptor)
            num_rows = flight_info.total_records
            num_cols = len(flight_info.schema)
            return (num_rows, num_cols)
        return self.data.shape

    @property
    def column_names(self):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars)
            }
            descriptor = pa.flight.FlightDescriptor.for_command(json.dumps(ticket))
            with ConnectionManager.get_connection() as conn:
                flight_info = conn.get_flight_info(descriptor)
            return flight_info.schema.names
        return self.data.column_names

    @property
    def total_bytes(self):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars)
            }
            descriptor = pa.flight.FlightDescriptor.for_command(json.dumps(ticket))
            with ConnectionManager.get_connection() as conn:
                flight_info = conn.get_flight_info(descriptor)
            return flight_info.total_bytes
        return self.data.nbytes

    def collect(self) -> DataFrame:
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars)
            }
            with ConnectionManager.get_connection() as conn:
                reader = conn.do_get(pa.flight.Ticket(json.dumps(ticket).encode('utf-8')))
            self.data = reader.read_all()
            self.actions = []
        return self

    def collect_blob(self) -> DataFrame:
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars),
                "type": "collect_blob"
            }
            with ConnectionManager.get_connection() as conn:
                reader = conn.do_get(pa.flight.Ticket(json.dumps(ticket).encode('utf-8')))
            self.data = reader.read_all()
            self.actions = []
        return self

    def get_stream(self, max_chunksize: Optional[int] = 1000):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars),
                "max_chunksize": max_chunksize
            }
            with ConnectionManager.get_connection() as conn:
                reader = conn.do_get(pa.flight.Ticket(json.dumps(ticket).encode('utf-8')))
            for batch in reader:
                yield batch.data
            self.actions = []
        else:
            for batch in self.data.to_batches(max_chunksize):
                yield batch

    def limit(self, rowNum: int) -> DataFrame:
        new_df = DataFrame(self.id, self.data, self.actions[:], self.connection_id)
        new_df.actions.append(("limit", {"rowNum": rowNum}))
        return new_df

    def slice(self, offset: int = 0, length: Optional[int] = None) -> DataFrame:
        new_df = DataFrame(self.id, self.data, self.actions[:], self.connection_id)
        new_df.actions.append(("slice", {"offset": offset, "length": length}))
        return new_df

    def select(self, *columns):
        new_df = DataFrame(self.id, self.data, self.actions[:], self.connection_id)
        new_df.actions.append(("select", {"columns": columns}))
        return new_df

    def filter(self, expression: str) -> DataFrame:
        new_df = DataFrame(self.id, self.data, self.actions[:], self.connection_id)
        new_df.actions.append(("filter", {"expression": expression}))
        return new_df

    def sum(self, column: str):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars),
                "column": column
            }
            with ConnectionManager.get_connection() as conn:
                results = conn.do_action(
                    pa.flight.Action("compute_sum", json.dumps(ticket).encode("utf-8")))
            for res in results:
                return json.loads(res.body.to_pybytes().decode("utf-8"))["result"]
        else:
            arrow_table = self.handle_prev_actions(self.data, self.actions)
            return pc.sum(arrow_table[column]).as_py()

    def mean(self, column):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars),
                "column": column
            }
            with ConnectionManager.get_connection() as conn:
                results = conn.do_action(
                    pa.flight.Action("compute_mean", json.dumps(ticket).encode("utf-8")))
            for res in results:
                return json.loads(res.body.to_pybytes().decode("utf-8"))["result"]
        else:
            arrow_table = self.handle_prev_actions(self.data, self.actions)
            return pc.mean(arrow_table[column]).as_py()

    def min(self, column):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars),
                "column": column
            }
            with ConnectionManager.get_connection() as conn:
                results = conn.do_action(
                    pa.flight.Action("compute_min", json.dumps(ticket).encode("utf-8")))
            for res in results:
                return json.loads(res.body.to_pybytes().decode("utf-8"))["result"]
        else:
            arrow_table = self.handle_prev_actions(self.data, self.actions)
            return pc.min(arrow_table[column]).as_py()

    def max(self, column):
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars),
                "column": column
            }
            with ConnectionManager.get_connection() as conn:
                results = conn.do_action(
                    pa.flight.Action("compute_max", json.dumps(ticket).encode("utf-8")))
            for res in results:
                return json.loads(res.body.to_pybytes().decode("utf-8"))["result"]
        else:
            arrow_table = self.handle_prev_actions(self.data, self.actions)
            return pc.max(arrow_table[column]).as_py()

    def sort(self, column: str, order: str = "ascending") -> DataFrame:
        new_df = DataFrame(self.id, self.data, self.actions[:], self.connection_id)
        new_df.actions.append(("sort", {"column": column, "order": order}))
        return new_df

    def sql(self, sql_str: str) -> DataFrame:
        new_df = DataFrame(self.id, self.data, self.actions[:], self.connection_id)
        new_df.actions.append(("sql", {"sql_str": sql_str}))
        return new_df

    def map(self, column: str, func: Any, new_column_name: Optional[str] = None) -> DataFrame:
        new_df = DataFrame(self.id, self.data, self.actions[:], self.connection_id)
        new_df.actions.append(("map", {"column": column, "func": func, "new_column_name": new_column_name}))
        return new_df

    def to_pandas(self, **kwargs) -> pandas.DataFrame:
        if self.data is None:
            with ConnectionManager.get_connection() as conn:
                reader = conn.do_get(pa.flight.Ticket(json.dumps(self, default=vars).encode('utf-8')))
            self.data = reader.read_all()
            self.actions = []
        return self.data.to_pandas(**kwargs)

    def to_pydict(self) -> Dict[str, List[Any]]:
        if self.data is None:
            with ConnectionManager.get_connection() as conn:
                reader = conn.do_get(
                    pa.flight.Ticket(json.dumps(self, default=vars).encode('utf-8')))
            self.data = reader.read_all()
            self.actions = []
        return self.data.to_pydict()

    def to_string(self, head_rows: int = 5, tail_rows: int = 5, first_cols: int = 3, last_cols: int = 3, display_all: bool = False) -> str:
        if self.data is None:
            ticket = {
                "dataframe": json.dumps(self, default=vars),
                "head_rows": head_rows,
                "tail_rows": tail_rows,
                "first_cols": first_cols,
                "last_cols": last_cols
            }
            with ConnectionManager.get_connection() as conn:
                results = conn.do_action(pa.flight.Action("to_string", json.dumps(ticket).encode("utf-8")))
                for res in results:
                    return res.body.to_pybytes().decode('utf-8')
        else:
            arrow_table = self.handle_prev_actions(self.data, self.actions)
            return format_arrow_table(arrow_table, head_rows, tail_rows, first_cols, last_cols, display_all)

    # todo: 补充
    def handle_prev_actions(self, arrow_table, prev_actions):
        for action in prev_actions:
            action_type, params = action
            if action_type == "limit":
                row_num = params.get("rowNum")
                arrow_table = arrow_table.slice(0, row_num)
            elif action_type == "slice":
                offset = params.get("offset", 0)
                length = params.get("length")
                arrow_table = arrow_table.slice(offset, length)
            elif action_type == "select":
                columns = params.get("columns")
                arrow_table = arrow_table.select(columns)
            elif action_type == "filter":
                column_names = arrow_table.column_names
                locals_dict = {col: arrow_table[col].to_pandas() for col in column_names}
                mask = eval(params.get("expression"), {"__builtins__": None}, locals_dict)
                arrow_table = arrow_table.filter(pa.array(mask))
            elif action_type == "map":
                column = params.get("column")
                func = params.get("func")
                new_column_name = params.get("new_column_name", f"{column}_mapped")
                column_data = arrow_table[column].to_pylist()
                mapped_data = [func(value) for value in column_data]
                arrow_table = arrow_table.append_column(new_column_name, pa.array(mapped_data))
            elif action_type == "sort":
                column = params.get("column")
                order = params.get("order", "ascending")
                if order == "ascending":
                    arrow_table = arrow_table.sort_by(column)
                else:
                    arrow_table = arrow_table.sort_by([(column, "descending")])
            elif action_type == "sql":
                dataframe = arrow_table
                arrow_table = duckdb.sql(params.get("sql_str")).arrow()
            else:
                raise ValueError(f"Unsupported action type: {action_type}")
        return arrow_table



from dfwriter.csv_writer import CSVWriter
from dfwriter.default_writer import DefaultWriter
from dfwriter.nc_writer import NCWriter
from dfwriter.tif_writer import TIFWriter


class DfWriter:
    # 1. 创建一个类级别的注册表，用于存储格式名称和处理类的映射
    _format_registry = {}

    def __init__(self):
        self.output_target = None
        self.output_format = "arrow"  # 默认格式

        # 2. 在初始化时注册内置/默认支持的格式
        self.register_default_formats()

    @classmethod
    def register_format(cls, format_name: str, writer_class):
        """
        插件注册方法：向 DfWriter 注册一个新的输出格式。
        这是一个类方法,例如 DfWriter.register_format(...)
        """
        cls._format_registry[format_name.lower()] = writer_class
        print(f"✅ Format '{format_name.lower()}' registered with writer {writer_class.__name__}")

    @classmethod
    def register_default_formats(cls):
        """一个辅助方法，用于注册所有内置格式，避免重复代码"""
        if not cls._format_registry:  # 只在第一次初始化时注册
            cls.register_format("csv", CSVWriter)
            cls.register_format("nc", NCWriter)
            cls.register_format("tif", TIFWriter)
            cls.register_format("tiff", TIFWriter)
            cls.register_format("arrow", DefaultWriter)

    def output(self, target):
        self.output_target = target
        return self

    def format(self, fmt):
        self.output_format = fmt.lower()
        return self

    def write(self, df):
        # 3. 从类级别的注册表中查找写入器
        writer_class = self._format_registry.get(self.output_format)

        if not writer_class:
            supported_formats = ", ".join(self._format_registry.keys())
            raise ValueError(
                f"Unsupported format: '{self.output_format}'. "
                f"Supported formats are: [{supported_formats}]"
            )

        # 实例化并使用写入器
        writer_instance = writer_class()

        # 先执行 collect() 确保数据在本地
        df.collect()
        arrow_table = df.data

        writer_instance.write(table=arrow_table, output_path=self.output_target)
        print(f"🚀 Successfully wrote data to '{self.output_target}' in '{self.output_format}' format.")